import os
import pytest
import rfmux
import time
import numpy as np
import matplotlib
matplotlib.use('Agg')  # Set the backend to 'Agg' for non-interactive use
import matplotlib.pyplot as plt
from scipy import signal
from scipy.stats import linregress
from sklearn.linear_model import LinearRegression
from sklearn.metrics import r2_score
from conftest import wait_for_action
'''

Welcome to the 1st draft of the CRS QC. To begin testing, type into the terminal 
$ pytest test_crs_integrated.py --serial=<serial> -s

If you wish to run only a set of particular tests:

$ pytest test_crs_integrated.py --serial=<serial> -k "<test_test_name_1> or <test_name_2>"

'''
# Ensure the board and the computer are connected to the same network.
# Most basic check, does not use hidfmux
def test_board_bootup(results):
    print("Retrieving board status")
    os.system("timeout 1 /home/joshua/code/crs-mkids/firmware/rfmux/r1.4/parser -i enp6s0 -t | grep 'Enclosure\\|Timestamp' > boot.txt")
    try:
        with open('boot.txt', 'r') as file:
            content = file.read()
    except FileNotFoundError:
        pytest.fail("No file was generated. Check your parser directory and connection to the board")
        
    assert content, "Error: No data received. The board might not be booted correctly."
    os.remove('boot.txt')
    print("Board boot status: OK")

    results['tests'].append({
        'title': 'Board Bootup Test',
        'status': 'OK'
        })
   
'''-------------------------------------HELPER FUNCTIONS---------------------------------------------'''
def transfer_function_helper(raw_sampl, AMPLITUDE, scale, attenuation):
    '''
    Calculates the dBm value of a sample from get_samples call.
    Predicts the dBm value of the signal from the parameters passed to it
    '''
    DB_LOSS_FACTOR = -2.56
    volts_peak_samples = (np.sqrt(2)) * np.sqrt(50 * (10 ** (-1.75 / 10)) / 1000) / 1880796.4604246316
    if np.iscomplexobj(raw_sampl):
        data_i = np.array(raw_sampl.real) * volts_peak_samples
        data_q = np.array(raw_sampl.imag) * volts_peak_samples
    else:
        data_i = np.array(raw_sampl.i) * volts_peak_samples
        data_q = np.array(raw_sampl.q) * volts_peak_samples
    magnitude = np.sqrt(data_i ** 2 + data_q ** 2)
    median_magnitude = np.median(magnitude)
    median_magnitude_dbm = 10 * np.log10(((median_magnitude / np.sqrt(2)) ** 2) / 50) + 30

    DAC_output_dBm = DB_LOSS_FACTOR + 10 * np.log10((AMPLITUDE ** 2) * 10 ** (scale / 10))
    predicted_magnitude_dbm = DAC_output_dBm - attenuation
    return median_magnitude_dbm, predicted_magnitude_dbm

def value_setter_helper(d, frequency, amplitude, scale, attenuation, channel, module, nco, target_dac, target_adc, highbank):
    '''

    NOTE: All the tests below are written with the following numbering conventions for highbanks
    set_amplitude, set_frequency, get_samples, set_nyqist_zone -- accept module 1-4
    set_scale, set_attenuation, set_nco -- accept module 1-8 
    
    '''

    d.clear_channels()
    d.set_frequency(frequency, d.UNITS.HZ, channel, module)
    d.set_amplitude(amplitude, d.UNITS.NORMALIZED, target_dac, channel, module)

    if highbank == True:
        module+=4
    d.set_dac_scale(scale, d.UNITS.DBM, module)
    d.set_adc_attenuator(attenuation, d.UNITS.DB, module=module)
    d.set_nco_frequency(nco, d.UNITS.HZ, target_dac, module=module)
    d.set_nco_frequency(-nco, d.UNITS.HZ, target_adc, module=module)

#small legacy helper used in checking how circular the i-q blob is
def covariance_matrix(samples):
    cov_matrix = np.cov(samples, rowvar=False)
    eigenvalues, _ = np.linalg.eigh(cov_matrix)
    ratio = eigenvalues[1] / eigenvalues[0]
    return ratio
'''------------------------------------------------------------------------------------------------------------'''
'''-------------------------------------------SENSORS--------------------------------------------------------------'''
def test_temperature(d, results, summary_results):
    TEMPERATURE_MIN = 0.0
    TEMPERATURE_MAX = 70.0
    errors = []
    test_status = 'Pass'

    sensors = (
        d.TEMPERATURE_SENSOR.MB_R5V0,
        d.TEMPERATURE_SENSOR.MB_R3V3A,
        d.TEMPERATURE_SENSOR.MB_R2V5,
        d.TEMPERATURE_SENSOR.MB_R1V8,
        d.TEMPERATURE_SENSOR.MB_R1V2A,
        d.TEMPERATURE_SENSOR.MB_R1V4,
        d.TEMPERATURE_SENSOR.MB_R1V2B,
        d.TEMPERATURE_SENSOR.MB_R0V85A, 
        d.TEMPERATURE_SENSOR.MB_R0V85B, 
        d.TEMPERATURE_SENSOR.RFSOC_PL, 
        d.TEMPERATURE_SENSOR.RFSOC_PS,
    )
    with d.tuber_context() as ctx:
        for s in sensors:
            ctx.get_motherboard_temperature(s)
        temps = ctx()
    for (s, v) in zip(sensors, temps):
        if not (TEMPERATURE_MIN <= v <= TEMPERATURE_MAX):
            errors.append(f"Sensor {s} with value {v} out of range {TEMPERATURE_MIN}, {TEMPERATURE_MAX}!")
            test_status = 'Fail'
        print(f"Temperature for sensor {s}: {v}")
    results['tests'].append({
        'title': 'Temperature Test',
        'sensors': [str(sensor) for sensor in sensors],
        'temperatures': temps,
        'errors': errors,
        'test_status': test_status
    })
    if 'Board Health' not in summary_results['summary']:
        summary_results['summary']['Board Health'] = {}
    summary_results['summary']['Board Health']['Temperature Test'] = test_status
    
def test_voltages(d, results,summary_results):

    VOLTAGE_TOL = 0.10  # 10%
    test_status = 'Pass'
    sensors = (
        (d.VOLTAGE_SENSOR.MB_VBP, 12),  # power supply
        (d.VOLTAGE_SENSOR.MB_R0V85A, 0.85),#
        (d.VOLTAGE_SENSOR.MB_R0V85B, 0.85),#
        (d.VOLTAGE_SENSOR.MB_R1V2A, 1.2),#
        (d.VOLTAGE_SENSOR.MB_R1V2B, 1.2),#
        (d.VOLTAGE_SENSOR.MB_R1V4, 1.4),#
        (d.VOLTAGE_SENSOR.MB_R1V8A, 1.8),#
        (d.VOLTAGE_SENSOR.MB_R2V5, 2.5),#
        (d.VOLTAGE_SENSOR.MB_R3V3A, 3.3),#
        (d.VOLTAGE_SENSOR.MB_R5V0, 5.0),#
        (d.VOLTAGE_SENSOR.RFSOC_PSMGTRAVCC, 0.9), #seen in in rfsoc_power
        (d.VOLTAGE_SENSOR.RFSOC_PSMGTRAVTT, 1.8), #CONFUSION top right in in rfsoc_power says 1.2 but acc output 1.8
        (d.VOLTAGE_SENSOR.RFSOC_VCCAMS, 1.8), # CONFUSION found VCCINTAMS - 0.85. but need a diff one
        (d.VOLTAGE_SENSOR.RFSOC_VCCAUX, 1.8), #in rfsoc_power
        (d.VOLTAGE_SENSOR.RFSOC_VCCBRAM, 0.85),# seen in rfsoc_power.sch
        (d.VOLTAGE_SENSOR.RFSOC_VCCINT, 0.85), #seen in rfsoc_power.sch
        (d.VOLTAGE_SENSOR.RFSOC_VCCPLAUX, 1.8), #
        (d.VOLTAGE_SENSOR.RFSOC_VCCPLINTFP, 0.85),#in rfsoc_power
        (d.VOLTAGE_SENSOR.RFSOC_VCCPLINTLP, 0.85),#in rfsoc_power
        (d.VOLTAGE_SENSOR.RFSOC_VCCPSAUX, 1.8),#in rfsoc_power
        (d.VOLTAGE_SENSOR.RFSOC_VCCPSDDR, 1.2), #
        (d.VOLTAGE_SENSOR.RFSOC_VCCPSINTFP, 0.85),#in rfsoc_power
        (d.VOLTAGE_SENSOR.RFSOC_VCCPSINTFPDDR, 0.85),#in rfsoc_power
        (d.VOLTAGE_SENSOR.RFSOC_VCCPSINTLP, 0.85), #in rfsoc_power
        (d.VOLTAGE_SENSOR.RFSOC_VCCPSIO0, 1.8),
        (d.VOLTAGE_SENSOR.RFSOC_VCCPSIO1, 3.3),
        (d.VOLTAGE_SENSOR.RFSOC_VCCPSIO2, 3.3),
        (d.VOLTAGE_SENSOR.RFSOC_VCCPSIO3, 3.3),
        (d.VOLTAGE_SENSOR.RFSOC_VCCVREFN, 0),
        (d.VOLTAGE_SENSOR.RFSOC_VCCVREFP, 1.85),
        (d.VOLTAGE_SENSOR.RFSOC_VCC_PSBATT, 0),#this is grounded as seen in rfsoc_power
        (d.VOLTAGE_SENSOR.RFSOC_VCC_PSDDRPLL, 1.8),# in rfsoc_power
        (d.VOLTAGE_SENSOR.RFSOC_VCC_PSPLL0, 1.2),# in rfsoc_power
    )
    #Get the real time data from sensors and check if it's within 10% of the allowed
    with d.tuber_context() as ctx:
        for (s, nv) in sensors:
            ctx.get_motherboard_voltage(s)
        voltages = ctx()
    for ((s, nv), v) in zip(sensors, voltages):
        min_v = nv * (1 - VOLTAGE_TOL)
        max_v = nv * (1 + VOLTAGE_TOL)

        if  round(v, 2) <min_v or round(v, 2)> max_v:
            print(f"Sensor {s} with value {v} out of range {min_v}, {max_v}!")
            test_status = 'Fail'
        print(f"Voltage for sensor {s}: {v}")

    results['tests'].append({
        'title': 'Voltage Test',
        'sensors': [str(sensor[0]) for sensor in sensors],
        'voltages': voltages,
        'test_status': test_status
    })

    if 'Board Health' not in summary_results['summary']:
        summary_results['summary']['Board Health'] = {}
    summary_results['summary']['Board Health']['Voltage Test'] = test_status

def test_currents(d, results, summary_results):
    CURRENT_LIMIT = 0.4  # 40% tolerance on the current limit
    test_status = 'Pass'
    sensors = (
        (d.CURRENT_SENSOR.MB_R0V85A, 2.5),
        (d.CURRENT_SENSOR.MB_R0V85B, 3.3),
        (d.CURRENT_SENSOR.MB_R1V2A, 2.4),
        (d.CURRENT_SENSOR.MB_R1V2B, 2.6),
        (d.CURRENT_SENSOR.MB_R1V4, 3.6),
        (d.CURRENT_SENSOR.MB_R1V8A, 3.3),
        (d.CURRENT_SENSOR.MB_R2V5, 4.1),
        (d.CURRENT_SENSOR.MB_R3V3A, 3.4),
        (d.CURRENT_SENSOR.MB_R5V0, 4.1),
        (d.CURRENT_SENSOR.MB_VBP, 4),
    )
    with d.tuber_context() as ctx:
        for (s, nc) in sensors:
            ctx.get_motherboard_current(s)
        currents = ctx()
    for ((s, nc), v) in zip(sensors, currents):
        min_c = nc * (1 - CURRENT_LIMIT)
        max_c = nc * (1 + CURRENT_LIMIT)
        if v > max_c:
            print(f"Sensor {s} with value {v} A above {max_c}!")
        if v < min_c:
            print(f"Sensor {s} with value {v} A below {min_c}!")

            test_status = 'Fail'
        print(f"Current for sensor {s}: {v}")
    results['tests'].append({
        'title': 'Current Test',
        'sensors': [str(sensor[0]) for sensor in sensors],
        'currents': currents,
        'test_status': test_status
    })
    if 'Board Health' not in summary_results['summary']:
        summary_results['summary']['Board Health'] = {}
    summary_results['summary']['Board Health']['Current Test'] = test_status
'''------------------------------------------------------------------------------------------------------------'''
'''-------------------------------------------DAC--------------------------------------------------------------'''
def test_dac_passband(d, results, summary_results, num_runs, test_highbank, results_dir):

    '''This test aims to determine 2 things: if the magnitude of the output signal from the DAC 
    is within expected margin and if the passband stays flat throughout the predicted frequency range (-300 -- 300) wrt the NCO.

    The algorithm takes the length of'frequencies' array to determine how many channels to set. If the frequencies array is >1000
    items, it records that repetitions of the process are needed (num_runs). NOTE: feel free to adjust the frequencies array to your 
    desired density. 
    Then, the 1000 channels are set to the amplitudes and frequencies -300 -- 300. Recorded sample magnitude is compared to the 
    predicted magnitude (predicted from the transfer functions). 
     
    0 Hz frequency is taken as reference. If recorded magnitude is >10% off from predicted -- fail. If the measured data on either side 
    of the 0Hz deviates by more than 3dB from center freuqency -- fail.
    
    AMPLITUDE - Intentionally chosed to be 0.001 to be able to set 1000 channels at once without exceeding the full scale
    NCO_FREQUENCY - Can be changed if desired. Was chosen as 625e6 to gauge module performance comparing it to old, measured data. 
    NOMINAL_AMPLITUDE, NOMINAL_FREQUENCY - are there to help set the attenuation, scale for each module, are completely nominal, can you any number.

    '''

    # FIXME: Some issue with the higher range of the frequencies showing a dip
    AMPLITUDE = 0.005 # FIXME: JM change by x5
    TARGET_DAC = d.TARGET.DAC
    TARGET_ADC = d.TARGET.ADC
    SAMPLES = 10
    NCO_FREQUENCY = 1e9+123
    ATTENUATION = 0
    SCALE = 7
    NOMINAL_AMPLITUDE = 0.5
    NOMINAL_FREQUENCY = 200e6
    NOMINAL_CHANNEL = 1
    TITLE = 'DAC Passband Test'

   #feel free to increase the range and density of frequencies
    frequencies = np.arange(-300e6, 300e6, 10e6)
    channels = np.arange(1, 1001)
    nruns = int(np.ceil(len(frequencies) / len(channels)))
    lenrun = min(len(frequencies), len(channels))
    test_result = {
        'title': TITLE,
        'modules': []
    }

    #always start with the lowbanks. This will only change to True (and algorithm will repeat) if the passed test_highbank ==True
    highbank_run = False

    #num_runs passed from the config file depending on whether or not there will be a highbank run
    for num_run in range(0, num_runs):
        for module in range(1, 5):

            #used for saving data in the dictionary under correct number (while being able to set values and get samples following the current conventions)
            module_actual = module
            
            if highbank_run:
                module_actual +=4

            print(f'DAC_passband - Testing Module #:{module_actual}')
            
            measured_dbm = []
            predicted_dbm = []
            module_errors = []
            for nrun in range(nruns):
                start_idx = nrun * lenrun
                end_idx = min(start_idx + lenrun, len(frequencies))
                d.clear_channels()
                time.sleep(0.2)

                #First, call the function to set all the parameters that won't change: attenuation, scale, nco
                value_setter_helper(d, NOMINAL_FREQUENCY,NOMINAL_AMPLITUDE, SCALE, ATTENUATION, NOMINAL_CHANNEL, module, NCO_FREQUENCY, TARGET_DAC, TARGET_ADC, highbank_run)
                
                with d.tuber_context() as ctx:
                    #Set all the nessesary channels for the run
                    for idx in range(start_idx, end_idx):
                        channel = int(channels[idx - start_idx])  # Convert to int and access correct channel
                        frequency = float(frequencies[idx])  # Convert to float
                        ctx.set_frequency(frequency, d.UNITS.HZ, channel, module)
                        ctx.set_amplitude(AMPLITUDE, d.UNITS.NORMALIZED, TARGET_DAC, channel, module)
                    ctx()
                time.sleep(0.2)

                # Get all the samples for the current run
                raw = d.py_get_samples(SAMPLES, channel=None, module=module)

                for idx in range(start_idx, end_idx):
                    channel = int(channels[idx - start_idx])  # Convert to int and access correct channel
                    rchan = np.array(raw.i[channel+1])+1j*np.array(raw.q[channel+1])
                    frequency = float(frequencies[idx])  # Ensure frequency is a float
                    measured, predicted = transfer_function_helper(rchan, AMPLITUDE, SCALE, ATTENUATION)
                    measured_dbm.append((frequency, measured))
                    predicted_dbm.append((frequency, predicted))

            mag_0 = next(meas for freq, meas in measured_dbm if freq == 0)
            
            #if 10% error from predicted -- fail
            if mag_0 < 1.1*predicted or mag_0 > 0.9*predicted:
                error = f"At a reference frequency (625MHz), generating a power output >10% off from predicted. Expected value is: {predicted}. Measured value was: {mag_0:.2f}"
                print(error)
                module_errors.append(error)

            #if 3dBm from predicted -- fail
            for freq, meas in measured_dbm:
                if abs(freq) <= 275e6 and meas <= mag_0 - 3:
                    error = f"Taking 0Hz as reference, there are significant gaps in the passband.Generating signals below -3dB point relative to 0Hz."
                    print(error)
                    module_errors.append(error)
                    break

            if not module_errors:
                print(f'Module {module_actual} passed the test')   

            module_result = {
                'module': module_actual,  # Ensure this is an int
                'frequencies': [freq/ 1e6 for freq, _ in measured_dbm],  # Convert frequencies to float
                'measured_dbm': [meas for _, meas in measured_dbm],  # Ensure this is a float
                'predicted_magnitude_dbm_values': [predicted for _, predicted in predicted_dbm],  # Convert to a list of floats
                'status': 'Fail' if module_errors else 'Pass',
                'error': '\n'.join(module_errors),
                'plot_path': 'plot_path'
            }

            #Determine the graph name / address for the plot (will be accessed in report_generator)
            graph_name = TITLE + "_module_"+ f'{module_result["module"]}' +'.png'
            plot_path = os.path.join(results_dir, 'plots', graph_name)
            module_result["plot_path"] = plot_path
            
            plt.figure(figsize=(10, 6))
            plt.plot(module_result['frequencies'], module_result['measured_dbm'], '-o', label = 'measured' )
            plt.plot(module_result['frequencies'], module_result['predicted_magnitude_dbm_values'], '-o', label = 'predicted' )
            plt.xlabel('Frequency (MHz)')
            plt.ylabel('Magnitude (dBm)')
            plt.legend()
            plt.title(f'Signal at DAC of module {module_result["module"]} with increasing frequency')
            plt.grid(True)
            plt.tight_layout()
            os.makedirs(os.path.dirname(plot_path), exist_ok=True)
            plt.savefig(plot_path)
            plt.close()
            

            test_result['modules'].append(module_result)

            if module_actual not in summary_results['summary']:
                summary_results['summary'][module_actual] = {}
            summary_results['summary'][module_actual]['Passband'] = module_result['status']

        #after 1 ru is complete, determine if a second run is needed
        if test_highbank:
            d.set_analog_bank(True)
            highbank_run = True
    
    d.set_analog_bank(False)
    results['tests'].append(test_result)


def test_dac_amplitude_transfer(d, results, summary_results, test_highbank, num_runs, results_dir):
    '''
    This test aims to determine if the output of the dac is scaled appropriately with the change in amplitude. The algorithm uses only 1 channel
    at each module a set frequency, varying its amplitude. The result is compared to the transfer function prediction
    
    '''
    SCALE = 1
    TARGET_DAC = d.TARGET.DAC
    TARGET_ADC = d.TARGET.ADC
    SAMPLES = 100
    ATTENUATION = 0
    CHANNEL = 1
    NCO_FREQUENCY = 0
    FREQUENCY = 150.5763e6
    NOMINAL_AMPLITUDE = 0
    TITLE = 'DAC Amplitude Transfer Test'

    test_result = {
        'title': TITLE,
        'modules': []
    }

    highbank_run = False
    #feel free to adjust the amplitude density as needed
    amplitudes = np.arange(0.05, 1, 0.05)
    d.clear_channels()

    #num_runs passed from the config file depending on whether or not there will be a highbank run
    for num_run in range(0, num_runs):
        for module in range(1, 5):
            module_errors = []
            amplitude_values = []
            magnitude_dbm_values = []
            predicted_magnitude_dbm_values = []

            if highbank_run:
                print(f'DAC_amplitude - Testing Module #:{module+4}')
            else:
                print(f'DAC_amplitude - Testing Module #:{module}')
            
            #When starting to test a new module, clear all channels and use the value_setter to set the variables that won't change attenuation, scale, frequency etc
            value_setter_helper(d, FREQUENCY, NOMINAL_AMPLITUDE, SCALE, ATTENUATION, CHANNEL, module, NCO_FREQUENCY, TARGET_DAC, TARGET_ADC, highbank_run)
            

            for amplitude in amplitudes:
                d.set_amplitude(amplitude, d.UNITS.NORMALIZED, TARGET_DAC, CHANNEL, module)
                samples = d.py_get_samples(SAMPLES, channel=CHANNEL, module=module)
                median_magnitude_dbm, predicted_magnitude_dbm = transfer_function_helper(samples, amplitude, SCALE, ATTENUATION)
                amplitude_values.append(amplitude)
                magnitude_dbm_values.append(median_magnitude_dbm)
                predicted_magnitude_dbm_values.append(predicted_magnitude_dbm)

            errors = np.abs(np.array(magnitude_dbm_values) - np.array(predicted_magnitude_dbm_values))
            if highbank_run:
                module +=4

            #Linearity check on the scaling: apply linear regression and check least square error
            log_amplitude_values = np.array(np.log10(amplitude_values)).reshape(-1, 1)
            regressor = LinearRegression().fit(log_amplitude_values, magnitude_dbm_values)
            linearity = r2_score(regressor.predict(log_amplitude_values), magnitude_dbm_values)


            #if any value is 3dB from predicted -- fail
            if np.any(errors >= 3):
                error = f'Generating a power output below 3dB of expected.'
                print(f'Module {module} {error}')
                module_errors.append(error)

            if linearity<0.999:
                error = 'The output is not linear on a lin-log scale.'
                print(error)
                module_errors.append(error)

            if not module_errors:
                print(f"Module {module} passed DAC amplitude test.")
            
            module_result = {
                'module': module,
                'amplitude_values': amplitude_values,
                'magnitude_dbm_values': magnitude_dbm_values,
                'predicted_magnitude_dbm_values': predicted_magnitude_dbm_values,
                'status': 'Fail' if module_errors else 'Pass',
                'error': '\n'.join(module_errors),
                'plot_path': 'plot_path'
            }

            graph_name = TITLE + "_module_"+ f'{module_result["module"]}' +'.png'
            plot_path = os.path.join(results_dir, 'plots', graph_name)
            module_result["plot_path"] = plot_path  

            plt.figure(figsize=(10, 6))
            plt.plot(module_result['amplitude_values'], module_result['magnitude_dbm_values'], '-o', label = 'measured' )
            plt.plot(module_result['amplitude_values'], module_result['predicted_magnitude_dbm_values'], '-o', label = 'predicted' )
            plt.xscale('log')
            plt.xlabel('Amplitude (Normalized)')
            plt.ylabel('Magnitude (dBm)')
            plt.legend()
            plt.title(f'DAC output of module {module_result["module"]} with increasing amplitude')
            plt.grid(True)
            plt.tight_layout()
            os.makedirs(os.path.dirname(plot_path), exist_ok=True)
            plt.savefig(plot_path)
            plt.close()

            test_result['modules'].append(module_result)
            if module not in summary_results['summary']:
                summary_results['summary'][module] = {}
            summary_results['summary'][module]['DAC Ampl'] = module_result['status']

        if test_highbank:
            d.set_analog_bank(True)
            highbank_run = True
        d.clear_channels()

    d.set_analog_bank(False)
    results['tests'].append(test_result)


def test_dac_scale_transfer(d, results, summary_results, test_highbank, num_runs, results_dir):
    '''
    This test aims to determine if the output of the dac is scaled appropriately with the change in 'scale' value (actually controlled 
    by varying the current in the balun). The algorithm uses only 1 channel at each module a set frequency, varying its scale. 
    The result is compared to the transfer function prediction

    Potentially add:
    Pick 4 channels, set the same frequency-> alter scale and plot the samples first, then convert to dBm
    plot together with prediction from the transfer funciton
    '''
    #set parameters for testing
    AMPLITUDE= 1
    NOMINAL_SCALE = 1
    TARGET_DAC = d.TARGET.DAC
    TARGET_ADC = d.TARGET.ADC
    SAMPLES = 100
    FREQUENCY = 150.5763e6
    NCO_FREQUENCY = 0
    ATTENUATION = 0
    CHANNEL = 1
    TITLE = 'DAC Scale Transfer Test'
    #dictionary just for this test will be appended to all the results at the end (could modify to avoid losing data if a pytest crash occurs)
    test_result = {
        'title': TITLE,
        'modules': []
    }
    for mod in range (1,4):
        d.set_adc_calibration_mode('MODE1', mod)

    highbank_run = False
    for num_run in range (0, num_runs):
        for module in range (1,5):
            # Initialize lists to store the scale and magnitude values
            module_errors = []
            scale_values = []
            magnitude_dbm_values = []
            predicted_magnitude_dbm_values = []

            d.clear_channels()
            value_setter_helper(d, FREQUENCY, AMPLITUDE, NOMINAL_SCALE, ATTENUATION, CHANNEL, module, NCO_FREQUENCY, TARGET_DAC, TARGET_ADC, highbank_run)
            module_sample = module

            if highbank_run:
                module_set= module +4
            else:
                module_set = module
            
            print(f'DAC_scale - Testing Module #: {module_set}')

            for scale in range (0, 8):
                
                d.set_dac_scale(scale, d.UNITS.DBM, module_set)
                raw_sampl = d.py_get_samples(SAMPLES, channel=CHANNEL, module=module_sample)
                median_magnitude_dbm, predicted_magnitude_dbm = transfer_function_helper(raw_sampl, AMPLITUDE, scale, ATTENUATION)
                
                scale_values.append(scale)
                magnitude_dbm_values.append(median_magnitude_dbm)
                predicted_magnitude_dbm_values.append(predicted_magnitude_dbm)

            #Power check on the DAC
            errors = np.abs(np.array(magnitude_dbm_values) - np.array(predicted_magnitude_dbm_values))
            
            if np.any(errors >= 3):
                error = f'Module {module_set} is generating a power output below 3dB of expected'
                print(error)
                module_errors.append(error)

            #Linearity check on the scaling: apply linear regression and check least square error
            scale_values = np.array(scale_values).reshape(-1, 1)
            regressor = LinearRegression().fit(scale_values, magnitude_dbm_values)
            linearity = r2_score(regressor.predict(scale_values), magnitude_dbm_values)
            if linearity<0.99:
                error = 'linearity check failed'
                module_errors.append(error)
                print(error)
            if not module_errors:
                print(f'Module {module_set} passed the DAC scaling test')
                    
            module_result = {
                'module': module_set,
                'scale_values': scale_values.flatten().tolist(),
                'magnitude_dbm_values': magnitude_dbm_values,
                'predicted_magnitude_dbm_values': predicted_magnitude_dbm_values,
                'status': 'Fail' if module_errors else 'Pass',
                'errors': f'{module_errors}',
                'plot_path': 'plot_path'
            }

            #same as in above tests, the path to the plot will be used by the report_generator
            graph_name = TITLE + "_module_"+ f'{module_result["module"]}' +'.png'
            plot_path = os.path.join(results_dir, 'plots', graph_name)
            module_result["plot_path"] = plot_path  

            plt.figure(figsize=(10, 6))
            plt.plot(module_result['scale_values'], module_result['magnitude_dbm_values'], '-o', label = 'measured' )
            plt.plot(module_result['scale_values'], module_result['predicted_magnitude_dbm_values'], '-o', label = 'predicted' )
            plt.xlabel('Scale(dBm)')
            plt.ylabel('Magnitude (dBm)')
            plt.legend()
            plt.title(f'DAC output of module {module_result["module"]} with increasing scale (dBm)')
            plt.grid(True)
            plt.tight_layout()
            os.makedirs(os.path.dirname(plot_path), exist_ok=True)
            plt.savefig(plot_path)
            plt.close()

            test_result['modules'].append(module_result)
            if module_set not in summary_results['summary']:
                summary_results['summary'][module_set] = {}
            summary_results['summary'][module_set]['DAC Scale'] = module_result['status']

        if test_highbank:
            d.set_analog_bank(True)
            highbank_run = True

    d.set_analog_bank(False)
    results['tests'].append(test_result)

def test_dac_mixmodes(d, results, summary_results, test_highbank):
    '''
    This test aims to verify the overall behaviour of the nyquist zones (1(NRZ) or 2(RC)) set in the firmware as well as a sanity check for the hardware response in
    the higher nquist regions. 
    The test uses 3 modules at a time, getting the behaviour of a singular nyquist region from a singular module. The report generator then will stitch the 
    data together into one graph per nyquist zone.
    '''

    MAX_TOTAL_FREQUENCY = 2.5e9  # in Hz
    DAC_MAX_FREQUENCY = 275e6  # in Hz (max DAC frequency)
    BANDWIDTH = 550e6  # in Hz (275 MHz on each side of NCO)
    AMPLITUDE = 1
    SCALE = 1
    TARGET_DAC = d.TARGET.DAC
    TARGET_ADC = d.TARGET.ADC
    SAMPLES = 100
    ATTENUATION = 0
    STEP = 50e6
    SAMPLING_FREQUENCY = 2.5e9  # in Hz
    MODULE_NYQUIST_MAP = {1: 1, 2: 2, 3: 3}
    CHANNEL = 1

    #obtained through a separate loopback measurement: transferfunction applied to a 1000-count get_samples call
    reference_data = {
            'mixmode_1': {
                'mod_region_1': {
                    'frequencies': [761.78e6, 1302.56e6, 1902.34e6],
                    'magnitudes': [-3.42, -4.82, -6.7]
                },
                'mod_region_2': {
                    'frequencies': [4338.22e6, 3697.44e6, 3097.66e6],
                    'magnitudes': [-24.87, -19.00, -13.82]
                },
                'mod_region_3': {
                    'frequencies': [5761.78e6, 6302.56e6, 6902.34e6],
                    'magnitudes': [-31.89, -27.84, -27.87]
                }
            },
            'mixmode_2': {
                'mod_region_1': {
                    'frequencies': [761.78e6, 1302.56e6, 1902.34e6],
                    'magnitudes': [-16.41, -12.24, -10.21]
                },
                'mod_region_2': {
                    'frequencies': [4238.22e6, 3697.44e6, 3097.66e6],
                    'magnitudes': [-12.30, -11.73,-10.01 ]
                },
                'mod_region_3': {
                    'frequencies': [5761.78e6, 6302.56e6, 6902.34e6],
                    'magnitudes': [-19.66, -20.60, -24.93]
                }
            }
        }

    def mixmode_test_for_bank(bank_name, is_highbank):
        test_result = {
            'title': f'DAC Mixmodes Test - {bank_name} bank',
            'analog_bank': bank_name,
            'mixmodes': []
        }

        d.set_analog_bank(is_highbank)

        for mixmode in range(1, 3):
            for module, nyquist_zone in MODULE_NYQUIST_MAP.items():
                module_setter = module
                if is_highbank:
                    module+=4

                module_errors = []
                module_magnitude_dbm_values = np.array([])
                frequencies = np.array([])
                print(f'Running module {module} in mixmode #{mixmode} nyquist region #{nyquist_zone} for {bank_name} bank')

                d.set_nyquist_zone(mixmode, module=module)
                nco = 0
                frequency = -DAC_MAX_FREQUENCY

                while nco < MAX_TOTAL_FREQUENCY:
                    while frequency <= DAC_MAX_FREQUENCY and nco + frequency < MAX_TOTAL_FREQUENCY:
                        if nco + frequency >= 0:
                            value_setter_helper(d, frequency, AMPLITUDE, SCALE, ATTENUATION, CHANNEL, module_setter, nco, TARGET_DAC, TARGET_ADC, is_highbank)
                            samples = d.py_get_samples(SAMPLES, channel=CHANNEL, module=module_setter)
                            module_magnitude_dbm, _ = transfer_function_helper(samples, AMPLITUDE, SCALE, ATTENUATION)

                            if nyquist_zone == 1:
                                frequencies = np.append(frequencies, frequency + nco)
                                module_magnitude_dbm_values = np.append(module_magnitude_dbm_values, module_magnitude_dbm)
                            elif nyquist_zone == 2:
                                frequencies = np.append(frequencies, 2 * SAMPLING_FREQUENCY - (frequency + nco))
                                module_magnitude_dbm_values = np.append(module_magnitude_dbm_values, module_magnitude_dbm)
                            elif nyquist_zone == 3:
                                frequencies = np.append(frequencies, 2 * SAMPLING_FREQUENCY + frequency + nco)
                                module_magnitude_dbm_values = np.append(module_magnitude_dbm_values, module_magnitude_dbm)
                        frequency += STEP
                    nco += BANDWIDTH
                    frequency = -DAC_MAX_FREQUENCY

                reference_freqs = np.array(reference_data[f'mixmode_{mixmode}'][f'mod_region_{nyquist_zone}']['frequencies'])
                reference_mags = np.array(reference_data[f'mixmode_{mixmode}'][f'mod_region_{nyquist_zone}']['magnitudes'])

                for reference_freq, reference_mag in zip(reference_freqs, reference_mags):
                    if len(frequencies) == 0:
                        module_errors.append(f"No frequencies measured for module {module} in Nyquist region {nyquist_zone}")
                        continue

                    closest_index = np.argmin(np.abs(frequencies - reference_freq))
                    obtained_magnitude = module_magnitude_dbm_values[closest_index]

                    if not (0.8 * abs(reference_mag) <= abs(obtained_magnitude) <= 1.15 * abs(reference_mag)):
                        module_errors.append(f"Frequency {frequencies[closest_index] / 1e6:.2f} MHz: measured {obtained_magnitude:.2f} dBm, expected {reference_freq/1e6:.2f} to have {reference_mag} dBm")
                
                if module_errors:
                    print(module_errors)
                else:
                    print('Module outputs within 15 percent of expected value')

                status = 'Fail' if module_errors else 'Pass'
                test_result['mixmodes'].append({
                    'mixmode': mixmode,
                    'module': module,
                    'frequencies': frequencies.tolist(),
                    'magnitude_dbm_values': module_magnitude_dbm_values.tolist(),
                    'reference_frequencies': reference_freqs.tolist(),
                    'reference_magnitudes': reference_mags.tolist(), 
                    'status': status
                })
                if module not in summary_results['summary']:
                    summary_results['summary'][module] = {}
                summary_results['summary'][module][f'Mixmode {mixmode}'] = status
                if mixmode == 2:
                    d.set_nyquist_zone(1, module=module)

        results['tests'].append(test_result)

    # Run test for low bank
    mixmode_test_for_bank('low', False)

    # If high bank needs to be tested, prompt the user to switch filters and run again
    if test_highbank:
        wait_for_action("Please switch filters from low to high banks before proceeding with high bank testing. \n \
                        1st Nyquist filter to ADC#5,\n\
                        2nd Nyquist filter to ADC#6,\n\
                        3rd Nyquist filter to ADC#7\n \
                        If already have filters connected - disregard this message")
        
        mixmode_test_for_bank('high', True)

    d.set_analog_bank(False)




'''---------------------------------------------------------------------------------------------------------------'''
'''--------------------------------------------------ADC----------------------------------------------------------'''

def test_adc_attenuation(d, results, summary_results, test_highbank, num_runs, results_dir):
    AMPLITUDE = 1
    TARGET_DAC = d.TARGET.DAC
    TARGET_ADC = d.TARGET.ADC
    SAMPLES = 100
    FREQUENCY = 80.234562e6
    NCO_FREQUENCY = 0
    SCALE = 1
    CHANNEL = 1
    ATTENUATION_0 = 0
    TITLE = 'ADC Attenuation Test'

    test_result = {
        'title': TITLE,
        'modules': []
    }

    highbank_run = False
    for num_run in range (0, num_runs):
        for module in range(1, 5):
            module_errors = []
            attenuation_values = []
            magnitude_dbm_values = []
            predicted_magnitude_dbm_values = []

            d.clear_channels()
            value_setter_helper(d, FREQUENCY, AMPLITUDE, SCALE, ATTENUATION_0, CHANNEL, module, NCO_FREQUENCY, TARGET_DAC, TARGET_ADC, highbank_run)

            module_sample = module
            if highbank_run:
                module_set= module +4
            else:
                module_set = module
            
            print(f'ADC_attenuation - Testing Module #: {module_set}')

            for attenuation in range(0, 15):
                d.set_adc_attenuator(attenuation, d.UNITS.DB, module=module_set)
                raw_sampl = d.py_get_samples(SAMPLES, channel=CHANNEL, module=module_sample)
                median_magnitude_dbm, predicted_magnitude_dbm = transfer_function_helper(raw_sampl, AMPLITUDE, SCALE, attenuation)
                attenuation_values.append(attenuation)
                magnitude_dbm_values.append(median_magnitude_dbm)
                predicted_magnitude_dbm_values.append(predicted_magnitude_dbm)

            attenuation_values = np.array(attenuation_values).reshape(-1, 1)
            regressor = LinearRegression().fit(attenuation_values, magnitude_dbm_values)
            linearity = r2_score(magnitude_dbm_values, regressor.predict(attenuation_values))

            if linearity < 0.99:
                error = f'linearity check failed'
                print(error)
                module_errors.append(error)

            if np.any(np.abs(np.array(magnitude_dbm_values) - np.array(predicted_magnitude_dbm_values)) >= 1):
                error = f'Module {module_set} is generating a power output lower than expected. Check your balun and SMA connections.'
                module_errors.append(error)
                print(error)

            else:
                print(f"Module {module_set} passed ADC attenuation tests.")
            module_result = {
                'module': module_set,
                'attenuation_values': attenuation_values.flatten().tolist(),
                'magnitude_dbm_values': magnitude_dbm_values,
                'predicted_magnitude_dbm_values': predicted_magnitude_dbm_values,
                'status': 'Fail' if module_errors else 'Pass',
                'errors': f'{module_errors}',
                'plot_path': 'plot_path'
            }

            graph_name = TITLE + "_module_"+ f'{module_result["module"]}' +'.png'
            plot_path = os.path.join(results_dir, 'plots', graph_name)
            module_result["plot_path"] = plot_path  

            plt.figure(figsize=(10, 6))
            plt.plot(module_result['attenuation_values'], module_result['magnitude_dbm_values'], '-o', label = 'measured' )
            plt.plot(module_result['attenuation_values'], module_result['predicted_magnitude_dbm_values'], '-o', label = 'predicted' )
            plt.xlabel('Attenuation(dBm)')
            plt.ylabel('Magnitude (dBm)')
            plt.legend()
            plt.title(f'Signal at ADC of module {module_result["module"]} with increasing attenuation')
            plt.grid(True)
            plt.tight_layout()
            os.makedirs(os.path.dirname(plot_path), exist_ok=True)
            plt.savefig(plot_path)
            plt.close()


            test_result['modules'].append(module_result)

            if module_set not in summary_results['summary']:
                summary_results['summary'][module_set] = {}
            summary_results['summary'][module_set][f'ADC Atten'] = module_result['status']   

        if test_highbank:
            d.set_analog_bank(True)
            highbank_run = True

    d.set_analog_bank(False)
    results['tests'].append(test_result)



@pytest.mark.skip(reason="no way of currently testing this")
def test_adc_even_phase(d, results):
    AMPLITUDE = 0.1
    TARGET_DAC = d.TARGET.DAC
    TARGET_ADC = d.TARGET.ADC
    SAMPLES = 100
    FREQUENCY = 1000
    NCO_FREQUENCY = 0
    SCALE = 1
    CHANNEL = 1
    module = 1
    test_result = {
        'title': 'ADC Even Phase Test',
        'modules': []
    }
    for module in range(1, 5):
        value_setter_helper(d, FREQUENCY, AMPLITUDE, SCALE, 0, CHANNEL, module, NCO_FREQUENCY, TARGET_DAC, TARGET_ADC)
        x = d.py_get_samples(SAMPLES, channel=1, module=1)
    
        data_i_samples = np.array(x.i)
        data_q_samples = np.array(x.q)
        data_samples = np.column_stack((data_i_samples, data_q_samples))
        ratio = covariance_matrix(data_samples)
        print(f"Eigenvalue ratio: {ratio}")
        module_result = {
            'module': module,
            'module': module,
            'data_i_samples': data_i_samples.tolist(),
            'data_q_samples': data_q_samples.tolist(),
            'ratio': ratio
        }
        test_result['modules'].append(module_result)
    
    results['tests'].append(test_result)
    
    if np.isclose(ratio, 1, atol=0.1):
        print("The cluster is approximately circular.")
    else:
        print("The cluster is more oval in shape.")

#used in the test_loopback_noise
def psd_helper(data_i_raw, data_q_raw):
    COMB_SAMPLING_FREQUENCY = 625e6
    volts_peak = (np.sqrt(2)) * np.sqrt(50 * (10 ** (-1.75 / 10)) / 1000) / 1880796.4604246316
    # Convert I and Q to peak volts
    tod_i = data_i_raw * volts_peak
    tod_q = data_q_raw * volts_peak

    # Define the sampling frequency for the ADC
    fir = (COMB_SAMPLING_FREQUENCY / 256 / 64 / 2 ** 6)
    # Apply Welch's method to get the signal PSD
    psdfreq, psd_i = signal.welch(tod_i/np.sqrt(4*50), fir, nperseg=len(tod_i// 4)) #tod_i / np.sqrt(4 * 50) 
    _, psd_q = signal.welch(tod_q/np.sqrt(4*50), fir, nperseg=len(tod_q// 4)) #tod_q / np.sqrt(4 * 50)

    # Convert from watts to dBm
    psd_i = 10 * np.log10(psd_i*1000)
    psd_q = 10 * np.log10(psd_q*1000)
    return psd_i, psd_q, psdfreq
       

def test_loopback_noise(d, results, summary_results, test_highbank, num_runs, results_dir):
    '''
    1. Send a signal with no amplitude to gauge the noise floor, and any spikes in the noise floor
    2. Send a higher power signal to dominate the output with 1/f noise. Determine the slope of 1/f and any unusual spikes
    '''
    TARGET_DAC = d.TARGET.DAC
    TARGET_ADC = d.TARGET.ADC
    SAMPLES = 10000

    # Set parameters for testing white noise floor
    AMPLITUDE_0 = 0
    FREQUENCY = 277.234562e6
    NCO_FREQUENCY = 0
    SCALE = 1
    ATTENUATION = 0
    CHANNEL_WHITE = 1

    # Set parameters for 1/f slope and peak noise
    AMPLITUDE_1_F = 1
    FREQUENCY_1_F = 277.987e6
    NCO_FREQUENCY_1_F = 0
    SCALE_1_F = 1
    ATTENUATION_1_F = 0
    CHANNEL_1_F = 784
    EXPECTED_SLOPE = -1
    TITLE = 'Loopback Noise Test'
    test_result = {
        'title': TITLE,
        'modules': []
    }
    highbank_run = False
    print(f'Just in case, i am not dead, just slow. At 10k samples -- 1 min/module. At 60k samples 4 min/module...')
    for num_run in range (0, num_runs):
        for mod in range(1,5):

            if highbank_run:
                print(f'Loopback Noise - Testing Module #:{mod+4}')
            else:
                print(f'Loopback Noise - Testing Module #:{mod}')

            floor_errors = []
            loud_errors = []

            # 1. Test the noise environment: send a signal with no amplitude and plot the PSD of the samples
            d.clear_channels()
            value_setter_helper(d, FREQUENCY, AMPLITUDE_0, SCALE, ATTENUATION, CHANNEL_WHITE, mod, NCO_FREQUENCY, TARGET_DAC, TARGET_ADC, highbank_run)
            raw_samples_white = d.py_get_samples(SAMPLES, channel=CHANNEL_WHITE, module=mod)
            data_i_white = np.array(raw_samples_white.i)
            data_q_white = np.array(raw_samples_white.q)
            psd_i_white, psd_q_white, psdfreq_white = psd_helper(data_i_white, data_q_white)


            # 2. Test the PSD of a large signal, establishing an expectation for 1/f slope and peak noise at low frequency
            value_setter_helper(d, FREQUENCY_1_F, AMPLITUDE_1_F, SCALE_1_F, ATTENUATION_1_F, CHANNEL_1_F, mod, NCO_FREQUENCY_1_F, TARGET_DAC, TARGET_ADC, highbank_run)
            raw_samples_psd = d.py_get_samples(SAMPLES, channel=CHANNEL_1_F, module=mod)
            data_i_1_f = np.array(raw_samples_psd.i)
            data_q_1_f = np.array(raw_samples_psd.q)
            psd_i_1_f, psd_q_1_f, psdfreq_1_f = psd_helper(data_i_1_f, data_q_1_f)


            # Compute the max points in the white noise and low frequency regions
            max_point_white_i = np.max(psd_i_white)
            max_point_white_q = np.max(psd_q_white)
            max_point_1_f_i = np.max(psd_i_1_f)
            max_point_1_f_q = np.max(psd_q_1_f)
            mean_point_white_i = np.mean(psd_i_white)
            mean_point_white_q = np.mean(psd_q_white)

            if highbank_run:
                mod +=4
            
            #check if any spikes are present
            if 0.95*(-163.4) < mean_point_white_i or 0.95*(-163.4) < mean_point_white_q:
                floor_errors.append('Noise floor of module {mod} is higher than expected')
            if  max_point_1_f_i > -79 or max_point_1_f_q >-79:
                loud_errors.append('Noise levels for frequencies dominated by 1/f are higher than -79dBm')
            

            # Filter the PSDs for plotting
            mask_plot = (psdfreq_1_f > 0) & (psdfreq_1_f < 150)
            psdfreq_1_f_plot = psdfreq_1_f[mask_plot]
            psd_i_1_f_plot = psd_i_1_f[mask_plot]
            psd_q_1_f_plot = psd_q_1_f[mask_plot]

            # Filter the PSDs for fitting
            mask_fit = (psdfreq_1_f > 0) & (psdfreq_1_f < 15)
            psdfreq_1_f_fit = psdfreq_1_f[mask_fit]
            psd_i_1_f_fit = psd_i_1_f[mask_fit] 
            psd_q_1_f_fit = psd_q_1_f[mask_fit]

            # Combine I and Q components for fitting
            combined_psd_1_f_fit = np.concatenate((psd_i_1_f_fit, psd_q_1_f_fit))
            combined_freqs_fit = np.concatenate((psdfreq_1_f_fit, psdfreq_1_f_fit))

            # Combine I and Q components for plotting
            combined_psd_1_f = np.concatenate((psd_i_1_f_plot, psd_q_1_f_plot))
            combined_freqs = np.concatenate((psdfreq_1_f_plot, psdfreq_1_f_plot))
            
            # Convert PSD from dBm/Hz to W/Hz for fitting
            psd_values_W_Hz_fit = 10 ** ((combined_psd_1_f_fit - 30) / 10)

            # Perform a linear fit in the log-log space
            log_frequencies_fit = np.log10(combined_freqs_fit)
            log_psd_values_fit = np.log10(psd_values_W_Hz_fit)

            # Linear regression to find the slope and intercept
            slope, intercept, r_value, p_value, std_err = linregress(log_frequencies_fit, log_psd_values_fit)

            # Calculate the fitted line in log space over the full range for plotting
            log_frequencies_plot = np.log10(combined_freqs)
            fitted_log_psd_values_plot = intercept + slope * log_frequencies_plot

            # Convert fitted values back to linear scale for plotting
            fitted_psd_values_W_Hz_plot = 10 ** fitted_log_psd_values_plot
            fitted_psd_values_dBm_Hz_plot = 10 * np.log10(fitted_psd_values_W_Hz_plot) + 30

            if slope < 1.1*EXPECTED_SLOPE or slope > 0.9*EXPECTED_SLOPE:
                loud_errors.append('1/f noise deviates from the expected value of -1')
        
            module_result = {
                'module': mod,
                'psd_i_white': psd_i_white.tolist(),
                'psd_q_white': psd_q_white.tolist(),
                'psd_frequencies_white': psdfreq_white.tolist(),
                'max_point_white_i': max_point_white_i.tolist(),
                'max_point_white_q': max_point_white_q.tolist(),
                'psd_i_1_f': psd_i_1_f.tolist(),
                'psd_q_1_f': psd_q_1_f.tolist(),
                'psd_frequencies_1_f': psdfreq_1_f.tolist(),
                'combined_psd_1_f': combined_psd_1_f.tolist(),
                'combined_freqs': combined_freqs.tolist(),  # This should be the original combined frequency list
                'max_point_1_f_i': max_point_1_f_i.tolist(),
                'max_point_1_f_q': max_point_1_f_q.tolist(),
                'fitted_psd_values_dBm_Hz': fitted_psd_values_dBm_Hz_plot.tolist(),
                'slope': slope,
                'white_status': 'Fail' if floor_errors else 'Pass',
                '1_f_status': 'Fail' if loud_errors else 'Pass',
                'plot_path_white': 'plot_path_white',
                'plot_path_1_f': 'plot_path_1_f'
            }

            graph_name_white = TITLE + "_white_"+"_module_"+ f'{module_result["module"]}' +'.png'
            plot_path_white = os.path.join(results_dir, 'plots', graph_name_white)
            module_result["plot_path_white"] = plot_path_white  

            graph_name_1_f = TITLE +"_1_f" + "_module_"+ f'{module_result["module"]}' +'.png'
            plot_path_1_f = os.path.join(results_dir, 'plots', graph_name_1_f)
            module_result["plot_path_1_f"] = plot_path_1_f

            #plot the white noise
            plt.figure(figsize=(10, 6))
            plt.plot(module_result['psd_frequencies_white'], module_result['psd_i_white'],  label = 'I' )
            plt.plot(module_result['psd_frequencies_white'], module_result['psd_q_white'], label = 'Q' )
            plt.scatter([module_result['psd_frequencies_white'][np.argmax(module_result['psd_i_white'])]], [module_result['max_point_white_i']], color='red', label=f'Max I spike: {module_result["max_point_white_i"]:.2f} dBm')
            plt.scatter([module_result['psd_frequencies_white'][np.argmax(module_result['psd_q_white'])]], [module_result['max_point_white_q']], color='green', label=f'Max Q spike: {module_result["max_point_white_q"]:.2f} dBm')
            plt.xlabel('Frequency (Hz from carrier)')
            plt.ylabel('Magnitude (dBm)')
            plt.xscale('log')
            plt.legend()
            plt.title(f'Noise floor in loopback at 277.23MHz at module {mod}')
            plt.grid(True)
            plt.tight_layout()
            os.makedirs(os.path.dirname(plot_path_white), exist_ok=True)
            plt.savefig(plot_path_white)
            plt.close()

             # 1/f noise PSD
            plt.figure(figsize=(10, 6))
            plt.plot(module_result['combined_freqs'], module_result['fitted_psd_values_dBm_Hz'], 'r', label=f'fit to 1/f spectrum: n = {module_result["slope"]:.2f}')
            plt.scatter(module_result['combined_freqs'], module_result['combined_psd_1_f'], label='1/f spectrum data')
            plt.scatter([module_result['psd_frequencies_1_f'][np.argmax(module_result['psd_i_1_f'])]], [module_result['max_point_1_f_i']], color='red', label=f'Max I spike: {module_result["max_point_1_f_i"]:.2f} dBm')
            plt.scatter([module_result['psd_frequencies_1_f'][np.argmax(module_result['psd_q_1_f'])]], [module_result['max_point_1_f_q']], color='green', label=f'Max Q spike: {module_result["max_point_1_f_q"]:.2f} dBm')
            plt.xscale('log')
            plt.title('PSD of 1/f Noise')
            plt.xlabel('Frequency (Hz)')
            plt.ylabel('PSD (dBm/Hz)')
            plt.legend()
            plt.tight_layout()
            os.makedirs(os.path.dirname(plot_path_1_f), exist_ok=True)
            plt.savefig(plot_path_1_f)
            plt.close()



            if mod not in summary_results['summary']:
                summary_results['summary'][mod] = {}
            summary_results['summary'][mod]['Noise floor'] = module_result['white_status']
            summary_results['summary'][mod]['1/f Noise'] = module_result['1_f_status']
        
            test_result['modules'].append(module_result)
        
        if test_highbank:
            d.set_analog_bank(True)
            highbank_run = True

    d.set_analog_bank(False)
    results['tests'].append(test_result)



def test_wideband_noise(d, results, summary_results, test_highbank, num_runs, results_dir):

    SAMPLING_FREQUENCY = 5e9  # 5 GHz
    NUM_SAMPLES = 4000  # Total number of samples
    highbank_run = False
    TITLE =  'Wideband FFT for Noise Detection'

    test_result = {
        'title': TITLE,
        'modules': []
    }

    for num_run in range(0, num_runs):
        for mod in range(1, 5):

            if highbank_run:
                print(f'Wideband FFT - Testing Module #:{mod + 4}')
            else:
                print(f'Wideband FFT - Testing Module #:{mod}')

            d.clear_channels()
            time.sleep(1)
            samples = d.get_fast_samples(NUM_SAMPLES, d.UNITS.RAW, module=mod)

            volts_peak_samples = (np.sqrt(2)) * np.sqrt(50 * (10 ** (-1.75 / 10)) / 1000) / 1880796.4604246316
            I_signal = np.array(samples.i) * volts_peak_samples
            Q_signal = np.array(samples.q) * volts_peak_samples


            psdfreq_i, psd_i = signal.welch(I_signal/np.sqrt(4*50), SAMPLING_FREQUENCY, nperseg=NUM_SAMPLES) #tod_i / np.sqrt(4 * 50) 
            psdfreq_q, psd_q = signal.welch(Q_signal/np.sqrt(4*50), SAMPLING_FREQUENCY, nperseg=NUM_SAMPLES) #tod_q / np.sqrt(4 * 50)

            # Convert from watts to dBm
            psd_i = 10 * np.log10(psd_i*1000)
            psd_q = 10 * np.log10(psd_q*1000)

            if highbank_run:
                mod+=4

            module_result = {
                'module': mod,
                'psd_i': psd_i.tolist(),
                'psd_q': psd_q.tolist(),
                'freqs_i': psdfreq_i.tolist(),
                'freqs_q': psdfreq_q.tolist(),
                'status': 'N/A',
                'plot_path': 'plot_path'
            }

            if mod not in summary_results['summary']:
                summary_results['summary'][mod] = {}
            summary_results['summary'][mod]['Wide FFT'] = module_result['status']

            graph_name = TITLE +"_module_"+ f'{module_result["module"]}' +'.png'
            plot_path = os.path.join(results_dir, 'plots', graph_name)
            module_result["plot_path"] = plot_path 

            #plot the white noise
            plt.figure(figsize=(10, 6))
            plt.plot(module_result['freqs_i'], module_result['psd_i'],  label = 'I' )
            plt.plot(module_result['freqs_q'], module_result['psd_q'], label = 'Q' )
            plt.xlabel('Frequency (Hz)')
            plt.ylabel('Magnitude (dBm/Hz)')
            plt.legend()
            plt.title(f'Wideband FFT at module {module_result["module"]}')
            plt.grid(True)
            plt.tight_layout()
            os.makedirs(os.path.dirname(plot_path), exist_ok=True)
            plt.savefig(plot_path)
            plt.close()

            test_result['modules'].append(module_result)

        if test_highbank:
            d.set_analog_bank(True)
            highbank_run = True

    d.set_analog_bank(False)
    results['tests'].append(test_result)
    



#!/usr/bin/env python3
"""
Simple script demonstrating the periscope algorithm control flow.
This script executes the complete measurement sequence:
1. Initialize
2. Network analysis
3. Unwrap cable delay
4. Find resonances
5. Take multisweep
6. Do fitting
7. Bias the kids
8. Get slow samples
9. Get PFB samples (only for real boards)

Usage:
    python simplified_tuning_flow.py MOCK         # Run with mock CRS
    python simplified_tuning_flow.py 0042         # Run with real CRS serial 0042
"""

import asyncio
import numpy as np
from datetime import datetime
import sys

# Import required modules
import rfmux
from rfmux.core.crs import CRS
from rfmux.core.transferfunctions import fit_cable_delay, calculate_new_cable_length
from rfmux.algorithms.measurement.fitting import find_resonances, fit_skewed_multisweep
from rfmux.algorithms.measurement.fitting_nonlinear import fit_nonlinear_iq_multisweep
from rfmux.algorithms.measurement.bias_kids import bias_kids

# Import mock mode support
from rfmux.mock.helpers import create_mock_crs


async def main(serial="MOCK"):
    """Main execution flow replicating periscope algorithm.
    
    Args:
        serial: CRS serial number (e.g., "0042") or "MOCK" for mock mode
    """
    
    # Configuration parameters
    MODULE = 1  # Use Module 1
    
    # Network analysis parameters
    NETANAL_PARAMS = {
        'amp': 0.001,
        'fmin': 0.6e9,      # 600 MHz
        'fmax': 1.1e9,     # 1100 MHz  
        'nsamps': 10,
        'npoints': 50000,
        'max_chans': 1023,
        'max_span': 500e6,  # 500 MHz
        'module': MODULE
    }
    
    # Find resonances parameters
    FIND_RES_PARAMS = {
        'min_dip_depth_db': 1.0,
        'min_Q': 1e4,
        'max_Q': 1e7,
        'min_resonance_separation_hz': 100e3,
        'data_exponent': 2.0
    }
    
    # Multisweep parameters
    MULTISWEEP_PARAMS = {
        'span_hz': 500e3,             # 5 MHz span around each resonance
        'npoints_per_sweep': 50,
        'amp': 0.001,
        'nsamps': 10,
        'module': MODULE,
        'bias_frequency_method': 'max-diq',  # Method to find optimal bias frequency
        'rotate_saved_data': False,          # Don't rotate data for df calibration consistency
        'sweep_direction': 'upward'
    }
    
    # Fitting parameters
    FIT_PARAMS = {
        'apply_skewed_fit': True,
        'apply_nonlinear_fit': True,  # Optional
        'approx_Q_for_fit': 1e4,
        'fit_resonances': True,
        'center_iq_circle': True,
        'normalize_fit': True
    }

    SAMPLE_PARAMS = {
        'num_samples': 1000,
        'return_spectrum': True,
        'scaling': 'psd',
        'reference': 'absolute',
        'nsegments': 5,
        'spectrum_cutoff': 0.9,
        'channel': None,
        'module': MODULE
    }
    
    is_mock = serial.upper() == "MOCK"
    
    print("="*60)
    print("Simple Periscope Algorithm Flow")
    print(f"Mode: {'MOCK' if is_mock else 'REAL HARDWARE'}")
    print(f"Serial: {serial}")
    print(f"Started at: {datetime.now()}")
    print("="*60)
    
    try:
        # Step 1: Initialize CRS connection
        print("\n1. Initializing CRS...")
        
        if is_mock:
            # Use mock CRS with simulated resonators
            crs = await create_mock_crs(
                module=MODULE,
                config={
                    'num_resonances': 10,  # More resonances for testing
                    'freq_start': 0.6e9,  # 600 MHz - within network analysis range
                    'freq_end': 1.0e9,    # 1 GHz - within network analysis range
                    'enable_bifurcation': False,  # Disable for simpler testing
                    'q_min': 5e3,  # Lower Q for easier detection
                    'q_max': 5e4,
                    'resonator_random_seed': 42 # making sure it is the same resonantor everytime
                },
                verbose=True
            )
            
            # For mock mode, we don't need to set timestamp port
            await crs.clear_channels(module=MODULE)
            print("   ✓ Mock CRS initialized with simulated resonators")
            
            # Run the algorithm flow with mock CRS
            await run_algorithm_flow(crs, MODULE, NETANAL_PARAMS, FIND_RES_PARAMS,
                                   MULTISWEEP_PARAMS, FIT_PARAMS, SAMPLE_PARAMS)
            
        else:
            # Use real hardware - load session with serial number
            session = rfmux.load_session(f'!HardwareMap [ !CRS {{ serial: "{serial}" }} ]')
            crs = session.query(CRS).one()
            
            # Resolve the connection
            await crs.resolve()
            print(f"   ✓ Connected to CRS serial: {serial}")
            
            # Set timestamp port
            if hasattr(crs, 'TIMESTAMP_PORT'):
                await crs.set_timestamp_port(crs.TIMESTAMP_PORT.TEST)
                print("   ✓ Timestamp port set to TEST")
            
            # Clear channels
            await crs.clear_channels(module=MODULE)
            print("   ✓ Channels cleared")
            
            # Run the algorithm flow
            await run_algorithm_flow(crs, MODULE, NETANAL_PARAMS, FIND_RES_PARAMS,
                                   MULTISWEEP_PARAMS, FIT_PARAMS, SAMPLE_PARAMS)
            
    except Exception as e:
        print(f"\n❌ Error occurred: {type(e).__name__}: {str(e)}")
        import traceback
        traceback.print_exc()
        return 1
    
    return 0


async def run_algorithm_flow(crs, MODULE, NETANAL_PARAMS, FIND_RES_PARAMS, 
                           MULTISWEEP_PARAMS, FIT_PARAMS, SAMPLE_PARAMS, full_run = True):
    """Run the complete algorithm flow with the given CRS instance."""
    
    # Step 2: Network Analysis
    print("\n2. Performing network analysis...")
    print(f"   Sweeping {NETANAL_PARAMS['fmin']/1e6:.1f} - {NETANAL_PARAMS['fmax']/1e6:.1f} MHz")
    
    # Progress callback for network analysis
    def netanal_progress_callback(module, percentage):
        print(f"   Network analysis progress: {percentage:.1f}%", end='\r')
    
    NETANAL_PARAMS['progress_callback'] = netanal_progress_callback
    
    netanal_result = await crs.take_netanal(**NETANAL_PARAMS)
    
    frequencies = netanal_result['frequencies']
    iq_complex = netanal_result['iq_complex']
    phase_degrees = netanal_result['phase_degrees']
    
    print(f"   ✓ Network analysis complete: {len(frequencies)} points")
    
    # Step 3: Unwrap cable delay
    print("\n3. Unwrapping cable delay...")
    
    # Fit cable delay from phase data
    tau_additional = fit_cable_delay(frequencies, phase_degrees)
    
    # Calculate new cable length (assume current is 0m for simplicity)
    current_cable_length = await crs.get_cable_length(module=MODULE)
    new_cable_length = calculate_new_cable_length(current_cable_length, tau_additional)
    
    # Set the cable length on the CRS
    await crs.set_cable_length(length=new_cable_length, module=MODULE)
    
    print(f"   ✓ Cable delay unwrapped: delay={tau_additional*1e9:.2f} ns")
    print(f"   ✓ Cable length set to: {new_cable_length:.2f} m")
    
    # Step 4: Find resonances
    print("\n4. Finding resonances...")
    
    resonance_result = find_resonances(
        frequencies=frequencies,
        iq_complex=iq_complex,
        **FIND_RES_PARAMS,
        module_identifier=f"Module {MODULE}"
    )
    
    resonance_frequencies = resonance_result['resonance_frequencies']
    resonance_details = resonance_result['resonances_details']

    if full_run:
        assert len(resonance_frequencies) == 10
    
    print(f"   ✓ Found {len(resonance_frequencies)} resonances:")
    for i, (freq, details) in enumerate(zip(resonance_frequencies, resonance_details)):
        print(f"     {i+1}: {freq/1e6:.3f} MHz, Q≈{details['q_estimated']:.0f}, depth={details['prominence_db']:.1f} dB")
    
    if not resonance_frequencies:
        print("   ! No resonances found. Adjust parameters and try again.")
        return
    
    # Step 5: Take multisweep
    print("\n5. Performing multisweep measurements...")
    print(f"   Sweeping around {len(resonance_frequencies)} resonances")
    
    # Set center frequencies for multisweep
    MULTISWEEP_PARAMS['center_frequencies'] = resonance_frequencies
    
    # Progress callback (optional)
    def progress_callback(module, percentage):
        print(f"   Progress: {percentage:.1f}%", end='\r')
    
    MULTISWEEP_PARAMS['progress_callback'] = progress_callback
    
    multisweep_results = await crs.multisweep(**MULTISWEEP_PARAMS)
    
    print(f"\n   ✓ Multisweep complete for {len(multisweep_results)} resonances")
    
    # Step 6: Do fitting
    print("\n6. Performing resonance fitting...")
    
    # Apply skewed Lorentzian fitting
    if FIT_PARAMS['apply_skewed_fit']:
        print("   Applying skewed Lorentzian fits...")
        multisweep_results = fit_skewed_multisweep(
            multisweep_results,
            approx_Q_for_fit=FIT_PARAMS['approx_Q_for_fit'],
            fit_resonances=FIT_PARAMS['fit_resonances'],
            center_iq_circle=FIT_PARAMS['center_iq_circle'],
            normalize_fit=FIT_PARAMS['normalize_fit']
        )
        
        # Count successful fits
        successful_fits = sum(1 for res_data in multisweep_results.values() 
                            if res_data.get('fit_params', {}).get('fr') != 'nan')
        print(f"   ✓ Skewed fitting complete: {successful_fits}/{len(multisweep_results)} successful")
    
    # Apply nonlinear fitting (optional)
    if FIT_PARAMS['apply_nonlinear_fit']:
        print("   Applying nonlinear fits...")
        multisweep_results = fit_nonlinear_iq_multisweep(
            multisweep_results,
            fit_nonlinearity=True,
            n_extrema_points=5,
            verbose=False
        )
        
        # Count successful nonlinear fits
        nl_successful = sum(1 for res_data in multisweep_results.values() 
                          if res_data.get('nonlinear_fit_success', False))
        print(f"   ✓ Nonlinear fitting complete: {nl_successful}/{len(multisweep_results)} successful")
    
    # Display fit results
    print("\n   Fit results summary:")
    for idx, res_data in multisweep_results.items():
        if isinstance(idx, (int, np.integer)):
            fit_params = res_data.get('fit_params', {})
            if fit_params.get('fr') != 'nan':
                print(f"     Resonance {idx}: fr={fit_params['fr']/1e6:.3f} MHz, "
                      f"Qr={fit_params['Qr']:.0f}, Qc={fit_params['Qc']:.0f}")
    
    # Step 7: Bias the KIDs
    print("\n7. Biasing the KIDs...")
    
    # Progress callback for bias_kids
    def bias_progress_callback(module, percentage):
        print(f"   Bias progress: {percentage:.1f}%", end='\r')
    
    # Pass multisweep results directly to bias_kids
    # Since we're doing a single amplitude sweep, we can pass the results directly
    bias_results = await bias_kids(
        crs=crs,
        multisweep_results=multisweep_results,
        module=MODULE,
        progress_callback=bias_progress_callback
    )
    
    print(f"\n   ✓ Bias complete for {len(bias_results)} detectors")
    
    # Display bias results
    print("\n   Bias results summary:")
    for det_idx, det_data in bias_results.items():
        if 'bias_frequency' in det_data:
            print(f"     Detector {det_idx}: bias_freq={det_data['bias_frequency']/1e6:.3f} MHz")
            if 'df_calibration' in det_data:
                print(f"                      df_cal={det_data['df_calibration']:.3e} Hz/rad")
    
    
    ### Step 8. Slow Noise spectrum 
    print("\n8. Collecting 1000 slow noise samples...")

    slow_data = await crs.py_get_samples(**SAMPLE_PARAMS)

    num_res = len(bias_results)

    print(f"\nSlow Noise data collected for the {num_res} biased channels")

    print("\n  Noise results summary:")

    for i in range(num_res):
        psd_i = slow_data.spectrum.psd_i[i]
        psd_q = slow_data.spectrum.psd_q[i]
        freq_iq = slow_data.spectrum.freq_iq
        freq_dsb = slow_data.spectrum.freq_dsb

        mean_psd_i = np.mean(psd_i[2:]) ### removing DC
        mean_psd_q = np.mean(psd_q[2:])

        print(f"   Channel {i+1}: Frequency Bandwidth = {max(freq_iq):.3f} Hz, Min dsb frequency = {min(freq_dsb):.3f} Hz, Max dsb frequency = {max(freq_dsb):.3f} Hz")
        print(f"                  Mean I spectrum power = {mean_psd_i:.3f} dBm/Hz, Mean Q spectrum power = {mean_psd_q:.3f} dBm/Hz\n")

    #### Step 9. PFB noise spectrum 
    if crs.serial == "0000":
        print("\nSkipping PFB noise samples for the mock mode")
    else:
        print("\n9. Collecting 100000 PFB noise samples....")

        for i in range(num_res):
            pfb_data = await crs.py_get_pfb_samples(100_000,
                                                    channel = i + 1,
                                                    module = 1,
                                                    binlim = 1e6,
                                                    trim = False,
                                                    nsegments = 5,
                                                    reference = "absolute",
                                                    reset_NCO = False)
            
            # print(f"Pfb data collected for channel {i+1}")
            
            pfb_psd_i = pfb_data.spectrum.psd_i
            pfb_psd_q = pfb_data.spectrum.psd_q
            pfb_freq_iq = pfb_data.spectrum.freq_iq
            pfb_freq_dsb = pfb_data.spectrum.freq_dsb

            mean_pfb_psd_i = np.mean(pfb_psd_i[2:]) ### Removing dc bin
            mean_pfb_psd_q = np.mean(pfb_psd_q[2:])

            print(f"   Channel {i+1}: Frequency Bandwidth = {max(pfb_freq_iq):.3f} Hz, Min dsb frequency = {min(pfb_freq_dsb):.3f} Hz, Max dsb frequency = {max(pfb_freq_dsb):.3f} Hz")
            print(f"                  Mean I spectrum power = {mean_pfb_psd_i:.3f} dBm/Hz, Mean Q spectrum power = {mean_pfb_psd_q:.3f} dBm/Hz\n")
            
    print("\n" + "="*60)
    print("Algorithm flow complete!")
    print(f"Finished at: {datetime.now()}")
    print("="*60)

if __name__ == "__main__":
    # Get serial number from command line or default to MOCK
    if len(sys.argv) > 1:
        serial = sys.argv[1]
    else:
        serial = "MOCK"
        print("No serial number provided, using MOCK mode")
    
    # Run the async main function
    exit_code = asyncio.run(main(serial=serial))
    exit(exit_code)



"""
Unified Startup Dialog for Periscope
====================================

This dialog is shown when Periscope launches, prompting the user to configure
both connection mode (Hardware/Mock/Offline) and session management (New/Load/None).
"""

from PyQt6 import QtWidgets, QtCore, QtGui
import datetime
from . import settings


class UnifiedStartupDialog(QtWidgets.QDialog):
    """
    Unified startup dialog for connection and session configuration.
    
    Presents connection modes:
    1. Serial, Module - Connect to hardware CRS
    2. Mock - Connect to mock CRS
    3. Offline - Browse session data without CRS connection
    
    And session modes (conditional on connection):
    1. New Session - Create a new session folder for auto-exporting data
    2. Load Session - Load a previously created session
    3. No Session - Skip session management (disable auto-export)
    
    Note: In Offline mode, only "Load Session" is available.
    """
    
    # Connection mode results
    CONN_HARDWARE = 1
    CONN_MOCK = 2
    CONN_OFFLINE = 3
    
    # Session mode results
    SESS_NEW = 1
    SESS_LOAD = 2
    SESS_NONE = 3
    
    def __init__(self, parent=None, prefill: dict | None = None):
        """
        Initialize the unified startup dialog.
        
        Args:
            parent: Parent widget (typically the main Periscope window)
            prefill: Optional dict with pre-fill values:
                - connection_mode: CONN_HARDWARE, CONN_MOCK, or CONN_OFFLINE
                - crs_serial: CRS serial number string
                - module: Module number (1-8)
        """
        super().__init__(parent)
        
        self.connection_mode = None
        self.session_mode = None
        self.crs_serial = None
        self.module = None
        self.session_path = None
        self.session_folder_name = None
        
        # Store pre-fill values
        self._prefill = prefill or {}
        
        self.setWindowTitle("Periscope Configuration")
        self.setModal(True)
        
        # Prevent closing via X button
        self.setWindowFlags(
            self.windowFlags() & ~QtCore.Qt.WindowType.WindowCloseButtonHint
        )
        
        self._setup_ui()
        
        # Apply pre-fill values after UI is set up
        self._apply_prefill()
        
    def _setup_ui(self):
        """Create the dialog UI."""
        layout = QtWidgets.QVBoxLayout(self)
        layout.setSpacing(20)
        layout.setContentsMargins(30, 30, 30, 30)
        
        # Title
        title_label = QtWidgets.QLabel("Welcome to Periscope")
        title_font = QtGui.QFont()
        title_font.setPointSize(16)
        title_font.setBold(True)
        title_label.setFont(title_font)
        title_label.setAlignment(QtCore.Qt.AlignmentFlag.AlignCenter)
        layout.addWidget(title_label)
        
        # Description
        desc_label = QtWidgets.QLabel(
            "Configure your connection and session settings to continue:"
        )
        desc_label.setWordWrap(True)
        desc_label.setAlignment(QtCore.Qt.AlignmentFlag.AlignCenter)
        layout.addWidget(desc_label)
        
        layout.addSpacing(10)
        
        # Connection section
        layout.addWidget(self._create_connection_section())
        
        # Session section
        layout.addWidget(self._create_session_section())
        
        # OK/Cancel buttons
        button_box = QtWidgets.QDialogButtonBox(
            QtWidgets.QDialogButtonBox.StandardButton.Ok |
            QtWidgets.QDialogButtonBox.StandardButton.Cancel
        )
        button_box.accepted.connect(self._validate_and_accept)
        button_box.rejected.connect(self.reject)
        layout.addWidget(button_box)
        
        # Set reasonable size
        self.setMinimumWidth(600)
        self.setMinimumHeight(600)
        
    def _create_connection_section(self) -> QtWidgets.QGroupBox:
        """Create the connection mode section."""
        group = QtWidgets.QGroupBox("Connection Mode")
        layout = QtWidgets.QVBoxLayout(group)
        
        # Connection radio buttons
        self.rb_hardware = QtWidgets.QRadioButton("Connect to Hardware (Serial, Module)")
        self.rb_hardware.setToolTip("Connect to a physical CRS board")
        self.rb_hardware.setChecked(True)  # Default
        
        self.rb_mock = QtWidgets.QRadioButton("Mock Mode")
        self.rb_mock.setToolTip("Use simulated CRS for testing")
        
        self.rb_offline = QtWidgets.QRadioButton("Offline Mode")
        self.rb_offline.setToolTip("Browse session data without CRS connection")
        
        layout.addWidget(self.rb_hardware)
        layout.addWidget(self.rb_mock)
        layout.addWidget(self.rb_offline)
        
        # Hardware connection details (Serial + Module)
        self.hardware_details = QtWidgets.QWidget()
        hardware_layout = QtWidgets.QFormLayout(self.hardware_details)
        hardware_layout.setContentsMargins(20, 10, 0, 10)
        
        self.serial_input = QtWidgets.QLineEdit()
        self.serial_input.setPlaceholderText("0042")
        self.serial_input.setToolTip("Serial number of the CRS board (e.g., 0042 for rfmux0042.local)")
        hardware_layout.addRow("CRS Serial:", self.serial_input)
        
        self.module_input = QtWidgets.QSpinBox()
        self.module_input.setRange(1, 8)
        self.module_input.setValue(1)
        self.module_input.setToolTip("Module number (1-8)")
        hardware_layout.addRow("Module:", self.module_input)
        
        layout.addWidget(self.hardware_details)
        
        # Connect signals to update UI state
        self.rb_hardware.toggled.connect(self._on_connection_mode_changed)
        self.rb_mock.toggled.connect(self._on_connection_mode_changed)
        self.rb_offline.toggled.connect(self._on_connection_mode_changed)
        
        # Initial state
        self._on_connection_mode_changed()
        
        return group
    
    def _create_session_section(self) -> QtWidgets.QGroupBox:
        """Create the session mode section."""
        group = QtWidgets.QGroupBox("Session Management")
        layout = QtWidgets.QVBoxLayout(group)
        
        # Session radio buttons
        self.rb_new_session = QtWidgets.QRadioButton("New Session")
        self.rb_new_session.setToolTip("Create a new session folder to auto-export data")
        self.rb_new_session.setChecked(True)  # Default
        
        self.rb_load_session = QtWidgets.QRadioButton("Load Existing Session")
        self.rb_load_session.setToolTip("Load a previously created session folder")
        
        self.rb_no_session = QtWidgets.QRadioButton("No Session")
        self.rb_no_session.setToolTip("Continue without session management (no auto-export)")
        
        layout.addWidget(self.rb_new_session)
        layout.addWidget(self.rb_load_session)
        layout.addWidget(self.rb_no_session)
        
        # New session details - just folder name input (path selected on OK click)
        self.new_session_details = QtWidgets.QWidget()
        new_session_layout = QtWidgets.QVBoxLayout(self.new_session_details)
        new_session_layout.setContentsMargins(20, 10, 0, 10)
        
        # Description
        new_desc = QtWidgets.QLabel("A folder selection dialog will open when you click OK.")
        new_desc.setStyleSheet("color: gray; font-style: italic;")
        new_session_layout.addWidget(new_desc)
        
        # Folder name input
        folder_name_layout = QtWidgets.QHBoxLayout()
        timestamp = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")
        self.folder_name_input = QtWidgets.QLineEdit()
        self.folder_name_input.setText(f"session_{timestamp}")
        self.folder_name_input.setToolTip("Name for the new session folder")
        folder_name_layout.addWidget(QtWidgets.QLabel("Folder Name:"))
        folder_name_layout.addWidget(self.folder_name_input)
        new_session_layout.addLayout(folder_name_layout)
        
        layout.addWidget(self.new_session_details)
        
        # Load session details - just description (path selected on OK click)
        self.load_session_details = QtWidgets.QWidget()
        load_session_layout = QtWidgets.QVBoxLayout(self.load_session_details)
        load_session_layout.setContentsMargins(20, 10, 0, 10)
        
        load_desc = QtWidgets.QLabel("A folder selection dialog will open when you click OK.")
        load_desc.setStyleSheet("color: gray; font-style: italic;")
        load_session_layout.addWidget(load_desc)
        
        layout.addWidget(self.load_session_details)
        
        # Connect signals
        self.rb_new_session.toggled.connect(self._on_session_mode_changed)
        self.rb_load_session.toggled.connect(self._on_session_mode_changed)
        self.rb_no_session.toggled.connect(self._on_session_mode_changed)
        
        # Initial state
        self._on_session_mode_changed()
        
        return group
    
    def _on_connection_mode_changed(self):
        """Handle connection mode radio button changes."""
        # Show/hide hardware details
        is_hardware = self.rb_hardware.isChecked()
        self.hardware_details.setVisible(is_hardware)
        
        # Enable/disable session modes based on connection
        is_offline = self.rb_offline.isChecked()
        
        # In offline mode, only "Load Session" is allowed
        # Only update if session widgets exist (they're created after connection widgets)
        if hasattr(self, 'rb_new_session'):
            self.rb_new_session.setEnabled(not is_offline)
        if hasattr(self, 'rb_no_session'):
            self.rb_no_session.setEnabled(not is_offline)
        
        # If switching to offline and not on "Load Session", switch to it
        if is_offline and hasattr(self, 'rb_load_session') and not self.rb_load_session.isChecked():
            self.rb_load_session.setChecked(True)
    
    def _on_session_mode_changed(self):
        """Handle session mode radio button changes."""
        is_new = self.rb_new_session.isChecked()
        is_load = self.rb_load_session.isChecked()
        
        self.new_session_details.setVisible(is_new)
        self.load_session_details.setVisible(is_load)
    
    def _validate_and_accept(self):
        """Validate inputs and open file dialogs as needed before accepting."""
        # Determine connection mode
        if self.rb_hardware.isChecked():
            self.connection_mode = self.CONN_HARDWARE
            # Validate serial number
            if not self.serial_input.text().strip():
                QtWidgets.QMessageBox.warning(
                    self,
                    "Missing CRS Serial",
                    "Please enter a CRS serial number for hardware connection."
                )
                return
            self.crs_serial = self.serial_input.text().strip()
            self.module = self.module_input.value()
            
        elif self.rb_mock.isChecked():
            self.connection_mode = self.CONN_MOCK
            # Mock mode defaults
            self.crs_serial = None  # Mock doesn't need serial
            self.module = self.module_input.value() if hasattr(self, 'module_input') else 1
            
        elif self.rb_offline.isChecked():
            self.connection_mode = self.CONN_OFFLINE
            self.crs_serial = None
            self.module = None
        
        # Get last session directory for file dialogs
        last_session_dir = settings.get_last_session_directory()
        
        # Determine session mode and open appropriate file dialogs
        if self.rb_new_session.isChecked():
            self.session_mode = self.SESS_NEW
            
            # Validate folder name first
            if not self.folder_name_input.text().strip():
                QtWidgets.QMessageBox.warning(
                    self,
                    "Missing Folder Name",
                    "Please enter a folder name for the new session."
                )
                return
            self.session_folder_name = self.folder_name_input.text().strip()
            
            # Open folder selection dialog for base path
            # Start from last used directory if available
            start_dir = last_session_dir if last_session_dir else ""
            base_path = QtWidgets.QFileDialog.getExistingDirectory(
                self,
                "Select Session Location",
                start_dir,
                QtWidgets.QFileDialog.Option.ShowDirsOnly | 
                QtWidgets.QFileDialog.Option.DontUseNativeDialog
            )
            
            # If user cancelled, stay in dialog
            if not base_path:
                return
            
            self.session_path = base_path
            # Save this directory for next time
            settings.set_last_session_directory(base_path)
            
        elif self.rb_load_session.isChecked():
            self.session_mode = self.SESS_LOAD
            
            # Open folder selection dialog for existing session
            # Start from last used directory if available
            start_dir = last_session_dir if last_session_dir else ""
            session_path = QtWidgets.QFileDialog.getExistingDirectory(
                self,
                "Select Session Folder",
                start_dir,
                QtWidgets.QFileDialog.Option.ShowDirsOnly |
                QtWidgets.QFileDialog.Option.DontUseNativeDialog
            )
            
            # If user cancelled, stay in dialog
            if not session_path:
                return
            
            self.session_path = session_path
            self.session_folder_name = None
            # Save the parent directory for next time (not the session folder itself)
            import os
            parent_dir = os.path.dirname(session_path)
            if parent_dir:
                settings.set_last_session_directory(parent_dir)
            
        elif self.rb_no_session.isChecked():
            self.session_mode = self.SESS_NONE
            self.session_path = None
            self.session_folder_name = None
        
        # Save connection settings for next time
        if self.connection_mode == self.CONN_HARDWARE:
            settings.set_last_connection_mode("hardware")
            if self.crs_serial:
                settings.set_last_crs_serial(self.crs_serial)
            if self.module:
                settings.set_last_module(self.module)
        elif self.connection_mode == self.CONN_MOCK:
            settings.set_last_connection_mode("mock")
            if self.module:
                settings.set_last_module(self.module)
        elif self.connection_mode == self.CONN_OFFLINE:
            settings.set_last_connection_mode("offline")
        
        # All validation passed
        self.accept()
    
    def _apply_prefill(self):
        """
        Apply pre-fill values from CLI arguments or saved settings to the UI.
        
        Priority: CLI prefill values > saved settings > defaults
        """
        # Load saved settings if no prefill provided
        if not self._prefill:
            # Try to restore from saved settings
            saved_mode = settings.get_last_connection_mode()
            saved_serial = settings.get_last_crs_serial()
            saved_module = settings.get_last_module()
            
            # Apply saved connection mode
            if saved_mode == "hardware":
                self.rb_hardware.setChecked(True)
            elif saved_mode == "mock":
                self.rb_mock.setChecked(True)
            elif saved_mode == "offline":
                self.rb_offline.setChecked(True)
            
            # Apply saved serial and module
            if saved_serial:
                self.serial_input.setText(saved_serial)
            if saved_module:
                self.module_input.setValue(saved_module)
        else:
            # Apply prefill values from CLI (takes precedence over saved settings)
            conn_mode = self._prefill.get('connection_mode')
            if conn_mode == self.CONN_HARDWARE:
                self.rb_hardware.setChecked(True)
            elif conn_mode == self.CONN_MOCK:
                self.rb_mock.setChecked(True)
            elif conn_mode == self.CONN_OFFLINE:
                self.rb_offline.setChecked(True)
            
            # Apply CRS serial
            crs_serial = self._prefill.get('crs_serial')
            if crs_serial:
                self.serial_input.setText(str(crs_serial))
            
            # Apply module
            module = self._prefill.get('module')
            if module is not None:
                self.module_input.setValue(int(module))
        
        # Trigger UI updates
        self._on_connection_mode_changed()
    
    def get_configuration(self) -> dict:
        """
        Get the selected configuration.
        
        Returns:
            dict: Configuration with keys:
                - connection_mode: CONN_HARDWARE, CONN_MOCK, or CONN_OFFLINE
                - session_mode: SESS_NEW, SESS_LOAD, or SESS_NONE
                - crs_serial: CRS serial number (for hardware) or None
                - module: Module number (for hardware/mock) or None
                - session_path: Base path (for new session) or full path (for load)
                - session_folder_name: Folder name (for new session) or None
        """
        return {
            'connection_mode': self.connection_mode,
            'session_mode': self.session_mode,
            'crs_serial': self.crs_serial,
            'module': self.module,
            'session_path': self.session_path,
            'session_folder_name': self.session_folder_name,
        }



"""
Objects model for the CRS
=========================

"""

__all__ = [
    "CRS",
    "Crate",
    "ReadoutModule",
    "ReadoutChannel",
    "Wafer",
    "Resonator",
    "ChannelMapping",
]

from .hardware_map import algorithm
from .schema import (
    CRS,
    ChannelMapping,
    Crate,
    ReadoutChannel,
    ReadoutModule,
    Resonator,
    Wafer,
)
from .session import HWMConstructor, HWMCSVConstructor, read_csv


from . import session

import sys
import os
import yaml


class YAMLLoader(session.YAMLLoader):

    # This is the default flavour
    flavour = sys.modules[__name__]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Plumbing
        self.add_constructor(u"!flavour", yaml_flavour_constructor)

        # Readout chain
        self.add_constructor(
            "!Crate",
            HWMConstructor(
                lambda loader: getattr(loader.flavour, "Crate"),
                AttributeMappingTouchup("slots", "slot"),
            ),
        )

        self.add_constructor(
            "!CRS",
            HWMConstructor(lambda loader: getattr(loader.flavour, "CRS")),
        )

        # Cryogenic chain
        self.add_constructor(
            "!Wafer", HWMConstructor(lambda loader: getattr(loader.flavour, "Wafer"))
        )

        self.add_constructor(
            "!Resonators",
            HWMCSVConstructor(lambda loader: getattr(loader.flavour, "Resonator")),
        )

        self.add_constructor("!ChannelMappings", ChannelMappingCSVConstructor)


def yaml_flavour_constructor(loader, node):
    """!flavour tag for YAML.

    This tag must occur within a HardwareMap directive. Note that it's
    swallowed by the HardwareMap, which ignores its children after
    instantiating them.  (The children add themselves in their instantiation
    methods.)
    """

    name = loader.construct_scalar(node)
    __import__(name)

    flavour = loader.flavour = sys.modules[name]

    if hasattr(flavour, "yaml_hook"):
        loader.register_finalization_hook(flavour.yaml_hook)

    return sys.modules[name]



def ChannelMappingCSVConstructor(loader, node):
    """
    Expect the following attributes:

    resonator:
        A path specifying a resonator (wafer/name or name)

    channel:
        The channel number used to index ReadoutChannel
    """

    fn = os.path.join(os.path.dirname(loader.name), node.value)
    dr = read_csv(fn)

    to_add = []

    for mapping in dr:
        # Parse 'readout_channel' entry into a ReadoutChannel mapping
        if "readout_channel" not in mapping:
            raise yaml.YAMLError("Missing 'readout_channel' mapping in ChannelMapping CSV")

        match mapping["readout_channel"].split("/"):
            case (serial, mod, channel):
                mapping["readout_channel"] = (
                    loader.hwm.query(ReadoutChannel)
                    .join(ReadoutModule)
                    .join(CRS)
                    .filter(
                        ReadoutChannel.channel == channel,
                        ReadoutModule.module == mod,
                        CRS.serial == serial,
                    )
                    .one()
                )

            case (crate, slot, mod, channel):
                mapping["readout_channel"] = (
                    loader.hwm.query(ReadoutChannel)
                    .join(ReadoutModule)
                    .join(CRS)
                    .join(Crate)
                    .filter(
                        ReadoutChannel.channel == channel,
                        ReadoutModule.module == mod,
                        CRS.slot == slot,
                        Crate.serial == crate,
                    )
                    .one()
                )

            case _:
                raise yaml.YAMLError(f"Unable to parse {mapping['readout_channel']} into a ReadoutChannel")

        # Parse 'resonator' entry into a Resonator mapping
        if "resonator" not in mapping:
            raise yaml.YAMLError("Missing 'resonator' mapping in ChannelMapping CSV")

        match mapping["resonator"].split("/"):
            case (wafer, resonator):
                mapping["resonator"] = (
                    loader.hwm.query(Resonator)
                    .join(Wafer)
                    .filter(Resonator.name == resonator, Wafer.name == wafer)
                    .one()
                )

            case _:
                raise yaml.YAMLError(f"Unable to parse {mapping['resonator']} into Resonator mapping!")

    loader.hwm.add_all([ChannelMapping(**x) for x in dr])


class AttributeMappingTouchup:
    """Correctly assign indexes for SQLAlchemy attribute_mapped_collections.

    HWM objects like ReadoutModule come with index columns like "module",
    which indicate their position in a collection (crs.modules) starting
    from 1. This idiom is convenient in ORM-land, but awkward to support in
    YAML serialization.
    """

    def __init__(self, group_attribute, member_attribute):
        self._group_attribute = group_attribute
        self._member_attribute = member_attribute

    def __call__(self, loader, mapping):

        if self._group_attribute in mapping:
            values = mapping[self._group_attribute]

            if isinstance(values, list):
                # We've been provided a list. Start numbering at 1.
                for index, value in enumerate(values):
                    if not value:
                        continue

                    # Don't clobber an existing numbering.
                    if getattr(value, self._member_attribute) is not None:
                        continue

                    setattr(value, self._member_attribute, index + 1)

                mapping[self._group_attribute] = [v for v in values if v]

            elif isinstance(values, dict):
                # We've been provided a dictionary. Assume the keys
                # provide the numbering.
                for key, value in values.items():
                    if not value:
                        continue
                    setattr(value, self._member_attribute, key)

                mapping[self._group_attribute] = [v for v in values.values() if v]

            else:
                raise TypeError(f"Expected a list, got '{values}'!")


# Tell session to use our YAMLLoader.
session.set_yaml_loader_class(YAMLLoader)

# vim: sts=4 ts=4 sw=4 tw=78 smarttab expandtab



"""
This module is derived from and based on the mr_resonator project by Maclean Rouble:
https://github.com/macleaner/mr_resonator

Modifications in this repository include:
- Trimmed to include only functions used by the Mock CRS framework
- Consolidation and integration with local JIT physics approximations

Original project license: see rfmux/mr_resonator/LICENSE (upstream LICENSE retained)
"""

import numpy as np

# Import Bessel approximations from JIT physics module
from .jit_physics import bessel_k0 as bessel_k0_approx, bessel_i0 as bessel_i0_approx

# Import MR_LEKID
from .mr_lekid import MR_LEKID as MR_LEKID


h = 6.626e-34      # Planck's constant [J·s]
kb = 1.38e-23      # Boltzmann constant [J/K]
mu0 = 4e-7 * 3.14159265359  # Permeability of free space [H/m] ≈ 1.257e-6


class MR_complex_resonator(): 
    """
    High-level KID physics model that maps material/geometry/operating-point
    parameters to a circuit-level LEKID (MR_LEKID) used by the mock framework.

    Overview
    --------
    - Computes complex surface impedance Zs from (σ1, σ2) using Gao-inspired
      approximations for superconducting films.
    - Converts Zs to total series resistance R and kinetic inductance Lk using
      geometry scaling (length/width).
    - Creates a MR_LEKID instance with (R, Lk, Lg, C, Cc, ...) and recenters
      the readout frequency via MR_LEKID.compute_fr() for a self-consistent
      operating point.
    - Provides calc_nqp() and related helpers to update base physics from
      quasiparticle number density and temperature.

    Notes
    -----
    - Units are explicitly called out in method docstrings (Hz, Ω, H, F, K).
    - Bessel functions K0/I0 are approximated via jit_physics for speed;
      relative error is typically ~1e-6 for the intended ranges.
    - MockResonatorModel reads attributes like readout_f, T, Delta0, N0,
      sigmaN, thickness, width, length, R_spoiler, and lekid for vectorized
      S21 calculations and convergence.
    """
    
    # Material constants for Aluminum (internal units: µm⁻³ J⁻¹ for N0, seconds for tau0)
    # N0 conversion: 1.72e10 µm⁻³ eV⁻¹ / (1.602e-19 J/eV) = 1.074e29 µm⁻³ J⁻¹
    AL_N0 = 1.72e10 / 1.602e-19     # Density of states [µm⁻³ J⁻¹]
    AL_TAU0 = 438e-9                # Electron-phonon time [s] (de Visser thesis)
    AL_TC = 1.2                     # Critical temperature [K]
    
    def __init__(self, T=0.12, base_readout_f=1e9, VL=540e-18, width=2e-6, thickness=30e-9, 
                 length=None, C=0.5e-12, Cc=0.005e-12, alpha_k=0.5, fix_Lg=None, R_spoiler=0, L_junk=0,
                 # Material parameters - Al defaults (can be overridden for custom materials)
                 Tc=None, N0=None, tau0=None, sigmaN=1./(4*20e-9),
                 # Operating point
                 Popt=1e-18, opt_eff=0.5, pb_eff=0.7, nu_opt=150e9, big_sigma_factor=1e-4, nstar=0,
                 # Readout chain
                 Vin=0.15e-3, input_atten_dB=20, ZLNA=50., GLNA=1,
                 # Deprecated parameter (kept for backward compatibility)
                 material=None,
                 verbose=False):
        """
        High-level KID physics model.
        
        Material Parameters
        -------------------
        Tc : float, optional
            Critical temperature [K]. Default: 1.2 K (Aluminum)
        N0 : float, optional
            Density of states at Fermi level [µm⁻³ J⁻¹]. 
            Default: 1.074e29 (Aluminum: 1.72e10 µm⁻³ eV⁻¹ converted to SI)
        tau0 : float, optional
            Characteristic electron-phonon time [s]. Default: 438 ns (Aluminum)
        sigmaN : float, optional
            Normal state conductivity [S/m]. Default: 1/(4*20nm) for thin Al
        """
        self.T = T
        self.readout_f = base_readout_f
        self.Popt = Popt
        self.opt_eff = opt_eff
        self.pb_eff = pb_eff
        self.nu_opt = nu_opt
        self.big_sigma_factor = big_sigma_factor
        
        # Handle deprecated material parameter
        if material is not None and material != 'Al':
            import warnings
            warnings.warn(
                f"material='{material}' is deprecated. Use explicit Tc, N0, tau0 parameters instead. "
                "Ignoring material name and using provided or default values.",
                DeprecationWarning
            )

        self.R_spoiler = R_spoiler
        self.L_junk = L_junk
               
        self.width = width
        self.thickness = thickness
        if length is not None:
            self.length = length
            VL = self.width * self.thickness * self.length
        else:
            self.length = VL / (self.width * self.thickness)
        self.VL = VL
        self.VL_um3 = VL*1e18 # in um^3; this is conventionally the units for nqp etc
        
        self.sigmaN = sigmaN
    
        # Material properties - use explicit values or Al defaults
        self.Tc = Tc if Tc is not None else self.AL_TC
        self.N0 = N0 if N0 is not None else self.AL_N0
        self.tau0 = tau0 if tau0 is not None else self.AL_TAU0
        
        if self.T >= self.Tc:
            raise ValueError('Error: cannot set operational temperature equal to or above transition temperature.')
        self.nstar = nstar


        self.Delta0 = 1.76 * kb * self.Tc
            
        # compute initial guess at resonator dark conductances and circuit values
        nqp = self.calc_nqp(T=T, Popt=self.Popt, opt_eff=opt_eff, pb_eff=pb_eff)
        self.sigma1_initial = self.calc_sigma1(f=self.readout_f, T=self.T, Popt=self.Popt)
        self.sigma2_initial = self.calc_sigma2(f=self.readout_f, T=self.T, Popt=self.Popt)
        self.sigma_initial = self.sigma1_initial - 1.j*self.sigma2_initial
        self.Zs_initial = self.calc_Zs(f=self.readout_f, sigma=self.sigma_initial)
        self.R_initial, self.Lk_initial = self.calc_R_L(f=self.readout_f, Zs=self.Zs_initial)
        # R_spoiler already included in calc_R_L(), no need to add again
        
        # generate dark resonator
        self.input_atten_dB = input_atten_dB
        self.C = C
        self.Cc = Cc
        self.Vin = Vin
        if fix_Lg is None:
            self.alpha_k = alpha_k
            self.Lg = (self.Lk_initial - self.alpha_k*self.Lk_initial) / self.alpha_k
        else:
            self.Lg = fix_Lg
            self.alpha_k = self.Lk_initial / (self.Lk_initial + self.Lg)
        
        self.lekid_params_initial = dict(R=self.R_initial, Lk=self.Lk_initial, Lg=self.Lg, C=self.C, Cc=self.Cc, Vin=self.Vin, input_atten_dB=self.input_atten_dB, ZLNA=ZLNA, GLNA=GLNA, L_junk=self.L_junk)
        if verbose:
            print('initial parameters:')
            print(self.lekid_params_initial)
        self.lekid_initial = MR_LEKID(**self.lekid_params_initial)
        
        if self.Lk_initial < 0:
            self.readout_f = base_readout_f
            if verbose:
                print('Warning: initial Lk guess is negative.')
        else:
#             self.readout_f = 1./np.sqrt(2*np.pi*(self.Lk_dark + self.Lg) * self.C)
            self.readout_f = self.lekid_initial.compute_fr()
            if verbose:
                print('base readout f: %.4e; readout f now: %.4e'%(base_readout_f, self.readout_f))
        
        # recompute the resonator parameters using the updated readout frequency:
        nqp = self.calc_nqp(T=T, Popt=self.Popt, opt_eff=opt_eff, pb_eff=pb_eff)
        self.sigma1_dark = self.calc_sigma1(f=self.readout_f, T=self.T, Popt=self.Popt)
        self.sigma2_dark = self.calc_sigma2(f=self.readout_f, T=self.T, Popt=self.Popt)
        self.sigma_dark = self.sigma1_dark - 1.j*self.sigma2_dark
        self.Zs_dark = self.calc_Zs(f=self.readout_f, sigma=self.sigma_dark)
        self.R_dark, self.Lk_dark = self.calc_R_L(f=self.readout_f, Zs=self.Zs_dark)
        # R_spoiler already included in calc_R_L(), no need to add again
        
        # generate dark resonator
        self.C = C
        self.Cc = Cc
        self.Vin = Vin
        if fix_Lg is None:
            self.alpha_k = alpha_k
            self.Lg = (self.Lk_dark - self.alpha_k*self.Lk_dark) / self.alpha_k
        else:
            self.Lg = fix_Lg
            self.alpha_k = self.Lk_dark / (self.Lk_dark + self.Lg)
        
        self.lekid_params_dark = dict(R=self.R_dark, Lk=self.Lk_dark, Lg=self.Lg, C=self.C, Cc=self.Cc, Vin=self.Vin, input_atten_dB=self.input_atten_dB, ZLNA=ZLNA, GLNA=GLNA, L_junk=self.L_junk)
        self.lekid = MR_LEKID(**self.lekid_params_dark, verbose=verbose)
        self.readout_f = self.lekid.compute_fr()
        if verbose:
            print(self.lekid_params_dark)

    

    def calc_Zs(self, f, sigma, thickness=None):#, sigma2=None):
        """
        Compute complex surface impedance from complex conductance.

        Parameters
        ----------
        f : float
            Probe frequency [Hz]
        sigma : complex
            Complex conductance σ = σ1 - j σ2 [S/m]
        thickness : float, optional
            Film thickness [m] (defaults to self.thickness)

        Returns
        -------
        complex
            Surface impedance Zs [Ω]
        """
        
        if thickness is None:
            thickness = self.thickness
#         print(thickness)
        root1 = (1.j*2*np.pi*f*mu0)/sigma
        cotharg = thickness * np.sqrt(1.j*2*np.pi*f*mu0*sigma)
        Zs = np.sqrt(root1) * 1./np.tanh(cotharg)
        return Zs


    def calc_R_L(self, f, Zs):
        """
        Convert surface impedance to total series R and kinetic inductance Lk.

        Parameters
        ----------
        f : float
            Probe frequency [Hz]
        Zs : complex
            Surface impedance [Ω]

        Returns
        -------
        tuple[float, float]
            (R_total [Ω], Lk_total [H])

        Notes
        -----
        Geometry scaling: multiply surface quantities by (length/width).
        R_spoiler is added as an extra series resistance term.
        """
        R = (Zs.real ) * (self.length / self.width) + self.R_spoiler
        L = (Zs.imag / (2*np.pi*f)) * (self.length / self.width)
        return R, L
    

    def zeta(self, f, T):
        """
        Dimensionless parameter ζ = h f / (2 k_B T).

        Parameters
        ----------
        f : float
            Frequency [Hz]
        T : float
            Temperature [K]

        Returns
        -------
        float
            ζ (dimensionless)
        """
        return h * f / (2 * kb * T)
    
    def calc_sigma1(self, f=None, nqp=None, T=None, Popt=None, pb_eff=None, opt_eff=None):
        """
        Real part of complex conductivity (σ1) following Gao (Eq. 2.96–2.97).

        Parameters
        ----------
        f : float, optional
            Probe frequency [Hz] (defaults to self.readout_f)
        nqp : float, optional
            Quasiparticle number density [um^-3] (defaults to calc_nqp())
        T : float, optional
            Temperature [K]
        Popt : float, optional
            Optical loading [W]
        pb_eff, opt_eff : float, optional
            Pair-breaking and optical efficiencies

        Returns
        -------
        float
            σ1 [S/m]

        Notes
        -----
        Uses fast bessel_k0 approximation from jit_physics; relative error ~1e-6
        over intended ranges.
        """
        
        if f is None:
            f = self.readout_f
        if T is None:
            T = self.T
        if Popt is None:
            Popt = self.Popt
        if opt_eff is None:
            opt_eff = self.opt_eff
        if pb_eff is None:
            pb_eff = self.pb_eff
        if nqp is None:
            nqp = self.calc_nqp(T=T, Popt=Popt, pb_eff=pb_eff, opt_eff=opt_eff)

        zeta = self.zeta(f=f, T=T)
        
        # Always use fast approximation (JIT-compiled)
        K0 = bessel_k0_approx(zeta)

        x1 = 2 * self.Delta0/(h*f)
        x2 = nqp / (self.N0 * np.sqrt(2*np.pi*kb*T*self.Delta0))

        return x1 * x2 * np.sinh(zeta) * K0 * self.sigmaN

    
    def calc_sigma2(self, f=None, nqp=None, T=None, Popt=None, pb_eff=None, opt_eff=None):
        """
        Imaginary part of complex conductivity (σ2) following Gao (Eq. 2.96–2.97).

        Parameters
        ----------
        f : float, optional
            Probe frequency [Hz] (defaults to self.readout_f)
        nqp : float, optional
            Quasiparticle number density [um^-3] (defaults to calc_nqp())
        T : float, optional
            Temperature [K]
        Popt : float, optional
            Optical loading [W]
        pb_eff, opt_eff : float, optional
            Pair-breaking and optical efficiencies

        Returns
        -------
        float
            σ2 [S/m]

        Notes
        -----
        Uses fast bessel_i0 approximation from jit_physics; relative error ~1e-6
        over intended ranges.
        """
        
        if f is None:
            f = self.readout_f
        if T is None:
            T = self.T
        if Popt is None:
            Popt = self.Popt
        if opt_eff is None:
            opt_eff = self.opt_eff
        if pb_eff is None:
            pb_eff = self.pb_eff
        if nqp is None:
            nqp = self.calc_nqp(T=T, Popt=Popt, pb_eff=pb_eff, opt_eff=opt_eff)
        Delta0 = self.Delta0
        
        zeta = self.zeta(f=f, T=T)
        
        # Always use fast approximation (JIT-compiled)
        I0 = bessel_i0_approx(zeta)

        x1 = np.pi * Delta0 / (h*f)
        x2 = nqp / (2*self.N0*Delta0)
        x3 = np.sqrt(2*Delta0/(np.pi*kb*T)) * np.exp(-zeta) * I0

        return x1 * (1 - x2*(1+x3)) * (self.sigmaN)

        
    def calc_nqp(self, T=None, Popt=None, opt_eff=None, pb_eff=None):
        """
        Quasiparticle number density from thermal + optical generation.

        Parameters
        ----------
        T : float, optional
            Temperature [K]
        Popt : float, optional
            Optical loading [W]
        opt_eff : float, optional
            Optical efficiency (0–1)
        pb_eff : float, optional
            Pair-breaking efficiency (0–1)

        Returns
        -------
        float
            nqp [um^-3]

        Notes
        -----
        nqp = sqrt(n_th^2 + (rate_optical / R)) - nstar
        where R depends on Delta(T), material constants, and tau0.
        """
        
        if T is None:
            T = self.T
        if Popt is None:
            Popt = self.Popt
        if opt_eff is None:
            opt_eff = self.opt_eff
        if pb_eff is None:
            pb_eff = self.pb_eff
            
        Delta = self.calc_Delta_gao(T=T)
        
        R = (2 * Delta)**2 / (2*self.N0 * self.tau0 * (kb * self.Tc)**3)
        
        nth = self.calc_nqp_th(T=T) + self.nstar
        rate_thermal = R * nth**2
        rate_optical = pb_eff * opt_eff * Popt / (Delta * self.VL_um3)
                
        return np.sqrt(nth**2 + rate_optical/R) - self.nstar

    def calc_nqp_th(self, T=None):
        """
        Thermal quasiparticle number density at temperature T.

        Parameters
        ----------
        T : float, optional
            Temperature [K]

        Returns
        -------
        float
            nqp_th [um^-3]

        Notes
        -----
        Uses the standard low-temperature approximation for superconductors.
        """
        
        if T is None:
            T = self.T
        Delta = self.calc_Delta_gao(T)
    
        nqp = 2 * self.N0 * np.sqrt(2 * np.pi * kb * T * Delta) * np.exp(-Delta / (kb * T))
        return nqp

    def calc_Delta_gao(self, T=None):
        """
        Superconducting gap energy Δ(T) via Gao's approximation.

        Parameters
        ----------
        T : float, optional
            Temperature [K]

        Returns
        -------
        float
            Δ(T) [J]

        Validity
        --------
        Representative up to ~0.7 Tc.

        Notes
        -----
        This is used to update R and Lk through conductivities and Zs.
        """
        if T is None:
            T = self.T
        
        innerexp = np.exp(-self.Delta0 / (kb*T))
        outerexp = np.exp( (-2*np.pi*kb * T / self.Delta0) * innerexp)
        return self.Delta0 * outerexp



"""
Tuber object interface
"""

from __future__ import annotations
import asyncio
import concurrent
import textwrap
import types
import warnings
import inspect

from . import TuberError, TuberStateError, TuberRemoteError
from .codecs import AcceptTypes, Codecs, TuberResult


__all__ = [
    "TuberObject",
    "SimpleTuberObject",
    "resolve",
    "resolve_simple",
]


async def resolve(hostname: str, objname: str | None = None, accept_types: list[str] | None = None):
    """Create a local reference to a networked resource.

    This is the recommended way to connect asynchronously to remote tuberd instances.
    """

    instance = TuberObject(objname, hostname=hostname, accept_types=accept_types)
    await instance.tuber_resolve()
    return instance


def resolve_simple(hostname: str, objname: str | None = None, accept_types: list[str] | None = None):
    """Create a local reference to a networked resource.

    This is the recommended way to connect serially to remote tuberd instances.
    """

    instance = SimpleTuberObject(objname, hostname=hostname, accept_types=accept_types)
    instance.tuber_resolve()
    return instance


def attribute_blacklisted(name: str):
    """
    Keep Python-specific attributes from being treated as potential remote
    resources. This blacklist covers SQLAlchemy, IPython, and Tuber internals.
    """

    if name.startswith(
        (
            "_sa",
            "_ipython",
            "_tuber",
        )
    ):
        return True

    return False


def tuber_wrapper(func: callable, name: str, meta: TuberResult):
    """
    Annotate the wrapper function with docstrings and signature.
    """

    docstring = ""

    # Begin with a function signature, if provided and valid
    if (sig := getattr(meta, "__signature__", None)) and isinstance(sig, str):
        docstring = f"{name}{sig}:\n\n"

    # Attach docstring, if provided and valid
    if (doc := getattr(meta, "__doc__", None)) and isinstance(doc, str):
        docstring += textwrap.dedent(meta.__doc__)

    func.__doc__ = docstring.strip()
    return func


def get_object_name(parent: str | list, attr: str | None = None, item: str | int | None = None):
    """
    Construct a valid object name for accessing objects in a registry.

    Arguments
    ---------
    parent : str
        Parent object name
    attr : str
        If supplied, this attribute name is joined with the parent name as "parent.attr".
    item : str or int
        If supplied, this item name is treated as an index into the parent as "parent[item]".

    Returns
    -------
    objname: list
        A valid object name.
    """
    if isinstance(parent, str):
        out = [parent]
    else:
        out = parent
    if item is None and attr is None:
        return out
    if attr is not None:
        out = out + [attr]
    if item is not None:
        last = out[-1]
        if isinstance(last, str):
            last = [last]
        else:
            last = list(last)
        out = out[:-1] + [tuple(last + [item])]
    return out


class SubContext:
    """A container for attributes of a Context object"""

    def __init__(self, objname: str, parent: "SimpleContext", attrname: str | None = None, **kwargs):
        self.objname = objname
        self.attrname = attrname
        self.parent = parent
        self.ctx_kwargs = kwargs
        self.container = {}

    def __call__(self, *args, **kwargs):
        """method-like sub-context"""
        kwargs.update(self.parent.ctx_kwargs)
        kwargs.update(self.ctx_kwargs)
        return self.parent._add_call(object=self.objname, method=self.attrname, args=args, kwargs=kwargs)

    def __getitem__(self, item: str | int):
        """container-like sub-context"""
        if item not in self.container:
            objname = get_object_name(self.objname, attr=self.attrname, item=item)
            self.container[item] = SubContext(objname, parent=self.parent)
        return self.container[item]

    def __getattr__(self, name: str):
        """object-like sub-context"""
        if attribute_blacklisted(name):
            raise AttributeError(f"{name} is not a valid method or property!")

        objname = get_object_name(self.objname, attr=self.attrname)
        caller = SubContext(objname, attrname=name, parent=self.parent)
        setattr(self, name, caller)
        return caller


class SimpleContext:
    """A serial context container for TuberCalls. Permits calls to be aggregated.

    Commands are dispatched strictly in-order, but are automatically bundled
    up to reduce roundtrips.
    """

    def __init__(self, obj: "SimpleTuberObject", *, accept_types: list[str] | None = None, **ctx_kwargs):
        self.calls: list[dict] = []
        self.obj = obj
        self.uri = f"http://{obj._tuber_host}/tuber"
        if accept_types is None:
            accept_types = self.obj._accept_types
        if accept_types is None:
            self.accept_types = list(AcceptTypes.keys())
        else:
            for accept_type in accept_types:
                if accept_type not in AcceptTypes.keys():
                    raise ValueError(f"Unsupported accept type: {accept_type}")
            self.accept_types = accept_types
        self.ctx_kwargs = ctx_kwargs
        self.container = {}

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.calls:
            self()

    def __getitem__(self, item: str | int):
        if item not in self.container:
            objname = get_object_name(self.obj._tuber_objname, item=item)
            self.container[item] = SubContext(objname, parent=self)
        return self.container[item]

    def __getattr__(self, name: str):
        if attribute_blacklisted(name):
            raise AttributeError(f"{name} is not a valid method or property!")

        # Queue methods of registry entries using the top-level registry context
        if self.obj._tuber_objname is None:
            ctx = SubContext([name], parent=self)
        else:
            ctx = SubContext(self.obj._tuber_objname, attrname=name, parent=self)

        setattr(self, name, ctx)
        return ctx

    def _add_call(self, **request):
        future = concurrent.futures.Future()
        self.calls.append((request, future))
        return future

    def send(self, continue_on_error: bool = False):
        """Break off a set of calls and return them for execution."""

        # An empty Context returns an empty list of calls
        if not self.calls:
            return

        calls = []
        futures = []
        while self.calls:
            (c, f) = self.calls.pop(0)

            calls.append(c)
            futures.append(f)

        if not hasattr(self.obj, "_tuber_session"):
            # session object should persist beyond the lifetime of the context,
            # akin to the asyncio event loop
            from requests_futures.sessions import FuturesSession

            self.obj._tuber_session = FuturesSession()

        cs = self.obj._tuber_session

        # Declare the media types we want to allow getting back
        headers = {"Accept": ", ".join(self.accept_types)}
        if continue_on_error:
            headers["X-Tuber-Options"] = "continue-on-error"

        # Create a HTTP request to complete the call.
        # Returns a Future whose result has been processed by the response hook.
        return cs.post(self.uri, json=calls, headers=headers, hooks={"response": self._response_hook(futures)})

    def _response_hook(self, futures: list["concurrent.futures.Future"]):
        """Hook function for parsing a response from the server into a list of futures for each context call"""

        def hook(r, *args, **kwargs):
            results = self._receive(r, futures)
            r.tuber_results = results
            return r

        return hook

    def _receive(self, response: "requests.Response", futures: list["concurrent.futures.Future"]):
        """Parse response from a previously sent HTTP request."""

        with response as resp:
            raw_out = resp.content
            if not resp.ok:
                try:
                    text = resp.text
                except Exception:
                    raise TuberRemoteError(f"Request failed with status {resp.status_code}")
                raise TuberRemoteError(f"Request failed with status {resp.status_code}: {text}")
            content_type = resp.headers["Content-Type"]
            # Check that the resulting media type is one which can actually be handled;
            # this is slightly more liberal than checking that it is really among those we declared
            if content_type not in AcceptTypes:
                raise TuberError(f"Unexpected response content type: {content_type}")
            json_out = AcceptTypes[content_type](raw_out, resp.apparent_encoding)

        if hasattr(json_out, "error"):
            # Oops - this is actually a server-side error that bubbles
            # through. (See test_tuberpy_async_context_with_unserializable.)
            # We made an array request, and received an object response
            # because of an exception-catching scope in the server. Do the
            # best we can.
            raise TuberRemoteError(json_out.error.message)

        for f, r in zip(futures, json_out):
            # Always emit warnings, if any occurred
            if hasattr(r, "warnings") and r.warnings:
                for w in r.warnings:
                    warnings.warn(w)

            # Resolve either a result or an error
            if hasattr(r, "error") and r.error:
                if hasattr(r.error, "message"):
                    f.set_exception(TuberRemoteError(r.error.message))
                else:
                    f.set_exception(TuberRemoteError("Unknown error"))
            else:
                if hasattr(r, "result"):
                    f.set_result(r.result)
                else:
                    f.set_exception(TuberError("Result has no 'result' attribute"))

        # Return a list of results
        return [f.result() for f in futures]

    def receive(self, response: "concurrent.futures.Future"):
        """Wait for a response from a previously sent HTTP request."""
        if response is None:
            return []
        return response.result().tuber_results

    def __call__(self, continue_on_error: bool = False):
        """Wait for any pending calls to complete and return the results from the server"""
        resp = self.send(continue_on_error=continue_on_error)
        return self.receive(resp)


class Context(SimpleContext):
    """An asynchronous context container for TuberCalls. Permits calls to be
    aggregated.

    Commands are dispatched strictly in-order, but are automatically bundled
    up to reduce roundtrips.
    """

    def __init__(self, obj: "TuberObject", **kwargs):
        super().__init__(obj, **kwargs)
        self.calls: list[tuple[dict, asyncio.Future]] = []

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """Ensure the context is flushed."""
        if self.calls:
            await self()

    def __enter__(self):
        raise NotImplementedError

    def __exit__(self):
        raise NotImplementedError

    def _add_call(self, **request):
        loop = asyncio.get_running_loop()
        future = loop.create_future()
        self.calls.append((request, future))
        return future

    async def __call__(self, continue_on_error: bool = False):
        """Break off a set of calls and return them for execution."""

        # An empty Context returns an empty list of calls
        if not self.calls:
            return []

        calls = []
        futures = []
        while self.calls:
            (c, f) = self.calls.pop(0)

            calls.append(c)
            futures.append(f)

        loop = asyncio.get_running_loop()
        if not hasattr(loop, "_tuber_session"):
            # hide import for non-library package that may not be invoked
            import aiohttp

            # AsyncResolver does not support mDNS
            resolver = aiohttp.resolver.ThreadedResolver()

            # Monkey-patch tuber session memory handling with the running event loop
            loop._tuber_session = aiohttp.ClientSession(
                json_serialize=Codecs["json"].encode, connector=aiohttp.TCPConnector(resolver=resolver)
            )

            # Ensure that ClientSession.close() is called when the loop is
            # closed.  ClientSession.__del__ does not close the session, so it
            # is not sufficient to simply attach the session to the loop to
            # ensure garbage collection.
            loop_close = loop.close

            def close(self):
                if hasattr(self, "_tuber_session"):
                    if not self.is_closed():
                        self.run_until_complete(self._tuber_session.close())
                    del self._tuber_session
                loop_close()

            loop.close = types.MethodType(close, loop)

        cs = loop._tuber_session

        # Declare the media types we want to allow getting back
        headers = {"Accept": ", ".join(self.accept_types)}
        if continue_on_error:
            headers["X-Tuber-Options"] = "continue-on-error"
        # Create a HTTP request to complete the call. This is a coroutine,
        # so we queue the call and then suspend execution (via 'yield')
        # until it's complete.
        async with cs.post(self.uri, json=calls, headers=headers) as resp:
            raw_out = await resp.read()
            if not resp.ok:
                try:
                    text = raw_out.decode(resp.charset or "utf-8")
                except Exception as ex:
                    raise TuberRemoteError(f"Request failed with status {resp.status}")
                raise TuberRemoteError(f"Request failed with status {resp.status}: {text}")
            content_type = resp.content_type
            # Check that the resulting media type is one which can actually be handled;
            # this is slightly more liberal than checking that it is really among those we declared
            if content_type not in AcceptTypes:
                raise TuberError("Unexpected response content type: " + content_type)
            json_out = AcceptTypes[content_type](raw_out, resp.charset)

        if hasattr(json_out, "error"):
            # Oops - this is actually a server-side error that bubbles
            # through. (See test_tuberpy_async_context_with_unserializable.)
            # We made an array request, and received an object response
            # because of an exception-catching scope in the server. Do the
            # best we can.
            raise TuberRemoteError(json_out.error.message)

        # Resolve futures
        for f, r in zip(futures, json_out):
            # Always emit warnings, if any occurred
            if hasattr(r, "warnings") and r.warnings:
                for w in r.warnings:
                    warnings.warn(w)

            # Resolve either a result or an error
            if hasattr(r, "error") and r.error:
                if hasattr(r.error, "message"):
                    f.set_exception(TuberRemoteError(r.error.message))
                else:
                    f.set_exception(TuberRemoteError("Unknown error"))
            else:
                if hasattr(r, "result"):
                    f.set_result(r.result)
                else:
                    f.set_exception(TuberError("Result has no 'result' attribute"))

        # Return a list of results
        return [await f for f in futures]


class SimpleTuberObject:
    """A base class for serial TuberObjects.

    This is a great way of using Python to correspond with network resources
    over a HTTP tunnel. It hides most of the gory details and makes your
    networked resource look and behave like a local Python object.

    To use it, you should subclass this SimpleTuberObject.
    """

    _context_class = SimpleContext
    _tuber_objname = None

    def __init__(
        self,
        objname: str | None,
        *,
        hostname: str | None = None,
        accept_types: list[str] | None = None,
        parent: "SimpleTuberObject" | None = None,
    ):
        self._tuber_objname = objname
        if parent is None:
            assert hostname, "Argument 'hostname' required"
            self._tuber_host = hostname
            self._accept_types = accept_types
        else:
            self._tuber_host = parent._tuber_host
            self._accept_types = parent._accept_types

    @property
    def is_container(self):
        """True if object is a container (list or dict) of remote items,
        otherwise False if resolved or None if not resolved."""
        if hasattr(self, "_tuber_meta"):
            return hasattr(self, "_items")

    def __getattr__(self, name: str):
        # Useful hint
        raise AttributeError(f"'{self._tuber_objname}' has no attribute '{name}'.  Did you run tuber_resolve()?")

    def __len__(self):
        try:
            return len(self._items)
        except AttributeError:
            raise TypeError(f"'{self._tuber_objname}' object has no len()")

    def __getitem__(self, item: str | int):
        try:
            return self._items[item]
        except AttributeError:
            raise TypeError(f"'{self._tuber_objname}' object is not subscriptable")

    def __iter__(self):
        try:
            return iter(self._items)
        except AttributeError:
            raise TypeError(f"'{self._tuber_objname}' object is not iterable")

    def object_factory(self, objname: str):
        """Construct a child TuberObject for the given resource name.

        Overload this method to create child objects using different subclasses.
        """
        return self.__class__(objname, parent=self)

    def tuber_context(self, **kwargs):
        """Return a context manager for aggregating method calls on this object."""

        return self._context_class(self, **kwargs)

    def tuber_resolve(self, force: bool = False):
        """Retrieve metadata associated with the remote network resource.

        This class retrieves object-wide metadata, which is used to build
        up properties and methods with tab-completion and docstrings.
        """
        if not force:
            try:
                return self._tuber_meta
            except AttributeError:
                pass

        with self.tuber_context() as ctx:
            ctx._add_call(object=self._tuber_objname, resolve=True)
            meta = ctx()
            meta = meta[0]

        return self._resolve_meta(meta)

    @staticmethod
    def _resolve_method(name: str, meta: TuberResult):
        """Resolve a remote method call into a callable function"""

        def invoke(self, *args, **kwargs):
            with self.tuber_context() as ctx:
                r = getattr(ctx, name)(*args, **kwargs)
            return r.result()

        return tuber_wrapper(invoke, name, meta)

    def _resolve_object(self, attr: str | None = None, item: str | int | None = None, meta: TuberResult | None = None):
        """Create a TuberObject representing the given attribute or container
        item, resolving any supplied metadata."""
        assert attr is not None or item is not None, "One of attr or item required"
        if self._tuber_objname is None:
            objname = [attr]
        else:
            objname = get_object_name(self._tuber_objname, attr=attr, item=item)
        obj = self.object_factory(objname)
        if meta is not None:
            obj._resolve_meta(meta)
        return obj

    def _resolve_meta(self, meta: TuberResult):
        """Parse metadata packet and recursively resolve all attributes."""

        # docstring
        if hasattr(meta, "__doc__"):
            self.__doc__ = meta.__doc__

        # object attributes
        for objname in getattr(meta, "objects", []) or []:
            objmeta = getattr(meta.objects, objname)
            obj = self._resolve_object(attr=objname, meta=objmeta)
            setattr(self, objname, obj)

        # methods
        if hasattr(meta, "methods"):
            # backwards compatibility for v0.15 and older: when assembling
            # metadata, older tuberd did not understand the "resolve=True"
            # argument and returned a list of methods as metadata, rather than
            # a dictionary that describes each one. Because this is a transient
            # workaround for newer client / older server installs, it's more
            # important that this code path is comprehensible than performant.
            # To get the job done, we create a transient SimpleTuberObject
            # rather than manage potentially async calls in what is normally a
            # synchronous context.
            if isinstance(meta.methods, list):
                sto = SimpleTuberObject(objname=self._tuber_objname, hostname=self._tuber_host)
                with sto.tuber_context() as ctx:
                    for m in meta.methods:
                        ctx._add_call(object=self._tuber_objname, property=m)
                    meta.methods = TuberResult(dict(zip(meta.methods, ctx())))

            for methname in meta.methods or {}:
                method = getattr(meta.methods, methname)
                if not callable(method):
                    method = self._resolve_method(methname, method)
                    # create method once and bind to each item in a container
                    setattr(meta.methods, methname, method)

                setattr(self, methname, types.MethodType(method, self))

        # static properties
        if hasattr(meta, "properties"):
            # same workaround as above
            if isinstance(meta.properties, list):
                sto = SimpleTuberObject(objname=self._tuber_objname, hostname=self._tuber_host)
                with sto.tuber_context() as ctx:
                    for m in meta.properties:
                        ctx._add_call(object=self._tuber_objname, property=m)
                    meta.properties = TuberResult(dict(zip(meta.properties, ctx())))

            for propname in meta.properties or {}:
                setattr(self, propname, getattr(meta.properties, propname))

        # container of objects
        if hasattr(meta, "values"):
            values = meta.values
            keys = getattr(meta, "keys", None)
            if keys is None or isinstance(keys, int):
                islist = True
                if isinstance(keys, int):
                    size = keys
                    values = [values] * size
                else:
                    size = len(values)
                keys = range(size)
                items = [None] * size
            else:
                islist = False
                if not isinstance(values, list):
                    values = [values] * len(keys)
                items = dict()

            for k, objmeta in zip(keys, values):
                items[k] = self._resolve_object(item=k, meta=objmeta)

            self._items = items

            if not islist:
                setattr(self, "keys", types.MethodType(lambda o: o._items.keys(), self))
                setattr(self, "values", types.MethodType(lambda o: o._items.values(), self))
                setattr(self, "items", types.MethodType(lambda o: o._items.items(), self))

            def tuber_get(self, name: str, keys: list[str | int] | None = None):
                """Get a property of every container item.

                Return a list of property values for each item.  If ``keys`` is
                supplied, return only the values corresponding to the given set
                of container items.
                """
                if keys is None:
                    if isinstance(self._items, list):
                        keys = range(len(self._items))
                    else:
                        keys = self._items.keys()
                return [getattr(self._items[k], name) for k in keys]

            setattr(self, "tuber_get", types.MethodType(tuber_get, self))

        self._tuber_meta = meta
        return meta


class TuberObject(SimpleTuberObject):
    """A base class for async TuberObjects.

    This is a great way of using Python to correspond with network resources
    over a HTTP tunnel. It hides most of the gory details and makes your
    networked resource look and behave like a local Python object.

    To use it, you should subclass this TuberObject.
    """

    _context_class = Context

    async def tuber_resolve(self, force: bool = False):
        """Retrieve metadata associated with the remote network resource.

        This class retrieves object-wide metadata, which is used to build
        up properties and methods with tab-completion and docstrings.
        """
        if not force:
            try:
                return self._tuber_meta
            except AttributeError:
                pass

        async with self.tuber_context() as ctx:
            ctx._add_call(object=self._tuber_objname, resolve=True)
            meta = await ctx()
            meta = meta[0]

        return self._resolve_meta(meta)

    @staticmethod
    def _resolve_method(name: str, meta: TuberResult):
        """Resolve a remote method call into an async callable function"""

        async def invoke(self, *args, **kwargs):
            async with self.tuber_context() as ctx:
                getattr(ctx, name)(*args, **kwargs)
                results = await ctx()
            return results[0]

        return tuber_wrapper(invoke, name, meta)


# vim: sts=4 ts=4 sw=4 tw=78 smarttab expandtab
